from __future__ import annotations

import logging
from collections import defaultdict

from angr.ailment.expression import Const
from angr.ailment.statement import ConditionalJump, Jump
from angr.analyses.decompiler.clinic import Clinic

from .edge import EdgeSort

log = logging.getLogger("utils.cfg")


def categorize_edges(disassembly, edges) -> None:
    """
    Categorize each edge.

    :param disassembly: A Disassembly analysis instance.
    :param list edges:  A list of edges.
    :return:            None
    """

    edges_by_node = defaultdict(list)

    for edge in edges:
        if edge.sort != EdgeSort.EXCEPTION_EDGE:
            edges_by_node[edge.src].append(edge)

    for src_node, items in edges_by_node.items():
        if len(items) == 1:
            # is it a back edge?
            # TODO: an accurate back edge identification requires us to identify loop heads. although we do identify
            # TODO: loop nodes at some point, the information is not available here...
            edge = items[0]
            if edge.src.addr >= edge.dst.addr:
                edge.sort = EdgeSort.BACK_EDGE

        elif len(items) == 2:
            # actually, let's determine which branch is the false branch
            edge_a, edge_b = items

            if isinstance(disassembly, Clinic):
                last_stmt = edge_a.src.statements[-1] if edge_a.src.statements else None
                fallthrough = None
                if isinstance(last_stmt, ConditionalJump):
                    fallthrough = last_stmt.false_target.value
                elif isinstance(last_stmt, Jump):
                    # this is unexpected since this block has two successors somehow, but the last statement is a Jump
                    # anyway we pick the current value as the fallthrough value
                    if isinstance(last_stmt.target, Const):
                        fallthrough = last_stmt.target.value
                else:
                    # a block whose last statement is not a jump or a conditional jump but has two successors?
                    fallthrough = None
            else:
                fallthrough = src_node.addr + src_node.size

            if edge_a.dst.addr == fallthrough and edge_b.dst.addr != fallthrough:
                edge_a.sort = EdgeSort.FALSE_BRANCH
                edge_b.sort = EdgeSort.TRUE_BRANCH
            elif edge_a.dst.addr != fallthrough and edge_b.dst.addr == fallthrough:
                edge_a.sort = EdgeSort.TRUE_BRANCH
                edge_b.sort = EdgeSort.FALSE_BRANCH
            else:
                # huh, there are either two fall-throughs or no fall-throughs
                pass



# pylint:disable=import-outside-toplevel,unused-argument
from __future__ import annotations

import subprocess
import sys
import threading
import time

import rpyc
from rpyc.utils.server import ThreadedServer

from angrmanagement.logic.singleton import SingleInstance, SingleInstanceException
from angrmanagement.utils.env import app_path

DEFAULT_PORT = 64000

CONNECTIONS = {}
TargetIDtoCONN = {}


class ManagementService(rpyc.Service):
    """
    Implements a binary management service. All daemon-enabled angr management will connect to this service and
    register the binary and target ID with this service.
    """

    _conn = None

    @staticmethod
    def _get_conn(target_id):
        if target_id in TargetIDtoCONN:
            conn = TargetIDtoCONN[target_id]
        else:
            raise Exception(
                f"The specified target {target_id} is not open in angr management. "
                f"We have the following ones: {str(TargetIDtoCONN)}."
            )
        return conn

    def on_connect(self, conn) -> None:
        self._conn = conn
        CONNECTIONS[conn] = None

    def on_disconnect(self, conn) -> None:
        self._conn = None
        if conn in CONNECTIONS:
            del CONNECTIONS[conn]

    def exposed_open(self, bin_path) -> None:
        if bin_path is None:
            return

        flags = {}
        if sys.platform.startswith("win"):
            DETACHED_PROCESS = 0x00000008
            flags["creationflags"] = DETACHED_PROCESS

        apppath = app_path(pythonw=False, as_list=True)
        shell = sys.platform.startswith("win")
        # default to using daemon
        # if the user chooses to use angr URL scheme to load a binary, they are more likely to keep interacting with
        # this binary using angr URL scheme, which requires the angr management instance to run in with-daemon mode.
        subprocess.Popen(
            apppath + ["-d", bin_path], shell=shell, stdin=None, stdout=None, stderr=None, close_fds=True, **flags
        )

    def exposed_jumpto(self, addr: int, symbol, target_id: str) -> None:
        conn = self._get_conn(target_id)
        conn.root.jumpto(addr, symbol)

    def exposed_register_client(self, target_id: str) -> None:
        TargetIDtoCONN[target_id] = self._conn

    def exposed_commentat(self, addr: int, comment: str, target_id: str) -> None:
        conn = self._get_conn(target_id)
        conn.root.commentat(addr, comment)

    def exposed_select_insns(self, addrs: list[int] | None, target_id: str):
        conn = self._get_conn(target_id)
        conn.root.select_insns(addrs)

    def exposed_exit(self) -> None:
        pass

    def exposed_custom_binary_aware_action(self, target_id: str, action, kwargs) -> None:
        conn = self._get_conn(target_id)
        conn.root.custom_binary_aware_action(action, kwargs)


def monitor_thread(server) -> None:
    """
    Monitors connection status, and kills the server (which is, the daemon process) after the last active connection is
    gone for 5 minutes.
    """

    last_active_conn = time.time()

    while True:
        if CONNECTIONS:
            # print("[*] Has %d active connections." % len(CONNECTIONS))
            last_active_conn = time.time()
        else:
            # print("[*] No active connection for %d seconds." % (time.time() - last_active_conn))
            pass

        if time.time() - last_active_conn > 300:
            # kill myself
            # print("[-] Shutting down the server.")
            server.close()
            break

        if server._closed:
            break

        time.sleep(1)


def start_daemon(port: int = DEFAULT_PORT) -> None:
    try:
        from angrmanagement.logic import GlobalInfo

        GlobalInfo.daemon_inst = SingleInstance()
    except SingleInstanceException:
        return

    # load plugins in headless mode
    from angrmanagement.plugins import PluginManager

    GlobalInfo.headless_plugin_manager = PluginManager(None)
    GlobalInfo.headless_plugin_manager.discover_and_initialize_plugins()

    # start the server
    server = ThreadedServer(ManagementService, port=port, protocol_config={"allow_public_attrs": True})
    threading.Thread(target=monitor_thread, args=(server,), daemon=True).start()
    server.start()


def daemon_exists() -> bool:
    try:
        inst = SingleInstance()
    except SingleInstanceException:
        return True
    del inst
    return False


def run_daemon_process() -> None:
    """
    Start a new Python process to run daemon.

    :return:
    """

    flags = {}
    if sys.platform.startswith("win"):
        DETACHED_PROCESS = 0x00000008
        flags["creationflags"] = DETACHED_PROCESS

    apppath = app_path(pythonw=True, as_list=True)
    subprocess.Popen(apppath + ["-D"], stdin=None, stdout=None, stderr=None, close_fds=True, **flags)


def daemon_conn(port: int = DEFAULT_PORT, service=rpyc.VoidService):
    return rpyc.connect("localhost", port, service=service, config={"allow_public_attrs": True})



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtGui import QKeySequence

from .menu import Menu, MenuEntry, MenuSeparator

if TYPE_CHECKING:
    from angrmanagement.ui.main_window import MainWindow


class AIMenu(Menu):
    """
    THe menu for AI-assisted features.
    """

    def __init__(self, main_window: MainWindow) -> None:
        super().__init__("A&I", parent=main_window)
        self.entries.extend(
            [
                MenuEntry(
                    "Refine &All",
                    main_window.llm_refine_all,
                    shortcut=QKeySequence("Ctrl+I"),
                ),
                MenuSeparator(),
                MenuEntry(
                    "Suggest &Variable Names",
                    main_window.llm_suggest_variable_names,
                ),
                MenuEntry(
                    "Suggest &Function Name",
                    main_window.llm_suggest_function_name,
                ),
                MenuEntry(
                    "Suggest Variable &Types",
                    main_window.llm_suggest_variable_types,
                ),
                MenuEntry(
                    "&Summarize Function",
                    main_window.llm_summarize_function,
                ),
            ],
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from archinfo.arch_arm import get_real_address_if_arm

from angrmanagement.data.object_container import ObjectContainer

if TYPE_CHECKING:
    from angr.sim_variable import SimVariable


class OperandHighlightMode:
    SAME_IDENT = 0
    SAME_TEXT = 1


class OperandDescriptor:
    __slots__ = (
        "text",
        "num_value",
        "func_addr",
        "variable_ident",
    )

    def __init__(self, text: str, num_value, func_addr=None, variable_ident=None) -> None:
        self.text = text
        self.num_value = num_value
        self.func_addr = func_addr
        self.variable_ident = variable_ident


class InfoDock:
    """
    Stores information associated to a disassembly view. Such information will be shared between the graph view and the
    linear view.
    """

    def __init__(self, disasm_view) -> None:
        super().__init__()
        self.disasm_view = disasm_view

        self.induction_variable_analysis = None
        self.variable_manager = None

        self.highlight_mode = OperandHighlightMode.SAME_IDENT  # default highlight mode

        self.selected_insns = ObjectContainer(set(), "The currently selected instructions")
        self.selected_operands = ObjectContainer({}, "The currently selected instruction operands")
        self.selected_blocks = ObjectContainer(set(), "The currently selected blocks")
        self.selected_ail_statements = ObjectContainer(set(), "The currently selected AIL statements")
        self.hovered_block = ObjectContainer(None, "The currently hovered block")
        self.hovered_edge = ObjectContainer(None, "The currently hovered edge")
        self.selected_labels = ObjectContainer(set(), "The currently selected labels")
        self.selected_variables = ObjectContainer(set(), "The currently selected variables")
        self.selected_block_tree_node = ObjectContainer(None, "The currently selected BlockTreeNode object")

    @property
    def smart_highlighting(self):
        return self.highlight_mode == OperandHighlightMode.SAME_IDENT

    @smart_highlighting.setter
    def smart_highlighting(self, v) -> None:
        if v:
            self.highlight_mode = OperandHighlightMode.SAME_IDENT
        else:
            self.highlight_mode = OperandHighlightMode.SAME_TEXT

    def initialize(self) -> None:
        self.selected_blocks.clear()
        self.selected_insns.clear()
        self.selected_operands.clear()
        self.hovered_block.am_obj = None

    def copy(self) -> InfoDock:
        r = InfoDock(self.disasm_view)
        r.variable_manager = self.variable_manager
        r.highlight_mode = self.highlight_mode
        r.selected_insns.am_obj = set(self.selected_insns.am_obj)
        r.selected_operands.am_obj = dict(self.selected_operands.am_obj)
        r.selected_blocks.am_obj = set(self.selected_blocks.am_obj)
        r.hovered_block.am_obj = self.hovered_block.am_obj
        r.hovered_edge.am_obj = self.hovered_edge.am_obj
        r.selected_labels.am_obj = set(self.selected_labels.am_obj)
        return r

    def hover_edge(self, src_addr, dst_addr) -> None:
        self.hovered_edge.am_obj = src_addr, dst_addr
        self.hovered_edge.am_event()

    def unhover_edge(self, src_addr, dst_addr) -> None:
        if self.hovered_edge.am_obj == (src_addr, dst_addr):
            self.hovered_edge.am_obj = None
            self.hovered_edge.am_event()

    def hover_block(self, block_addr) -> None:
        self.hovered_block.am_obj = block_addr
        self.hovered_block.am_event()

    def unhover_block(self, block_addr) -> None:
        if self.hovered_block.am_obj == block_addr:
            self.hovered_block.am_obj = None
            self.hovered_block.am_event()

    def select_block(self, block_addr) -> None:
        self.selected_blocks.clear()  # selecting one block at a time
        self.selected_blocks.add(block_addr)
        self.selected_blocks.am_event()
        self._update_published_view_state()

    def select_instruction(self, insn_addr, unique: bool = True, insn_pos=None, use_animation: bool = True) -> None:
        arch = self.disasm_view.instance.project.arch
        real_addr = get_real_address_if_arm(arch, insn_addr)
        self.disasm_view.set_synchronized_cursor_address(real_addr)

        self.unselect_all_labels()
        self.unselect_block_tree_node()
        if insn_addr not in self.selected_insns:
            if unique:
                # unselect existing ones
                self.unselect_all_instructions()
                self.selected_insns.add(insn_addr)
            else:
                self.selected_insns.add(insn_addr)
            self.disasm_view.current_graph.show_instruction(insn_addr, insn_pos=insn_pos, use_animation=use_animation)
            self.selected_insns.am_event(insn_addr=insn_addr)

        self._update_published_view_state()

    def unselect_instruction(self, insn_addr) -> None:
        if insn_addr in self.selected_insns:
            self.selected_insns.remove(insn_addr)
            self.selected_insns.am_event()
        self._update_published_view_state()

    def unselect_all_instructions(self) -> None:
        if self.selected_insns:
            self.selected_insns.clear()
            self.selected_insns.am_event()
        self._update_published_view_state()

    def select_ail_statement(self, stmt_addr, unique: bool = True) -> None:
        self.disasm_view.set_synchronized_cursor_address(stmt_addr[0])

        self.unselect_all_labels()
        self.unselect_all_instructions()
        if stmt_addr not in self.selected_ail_statements:
            if unique:
                # unselect existing ones
                self.unselect_all_instructions()
                self.selected_ail_statements.add(stmt_addr)
            else:
                self.selected_ail_statements.add(stmt_addr)
            self.selected_ail_statements.am_event(stmt_addr=stmt_addr)

        self._update_published_view_state()

    def unselect_ail_statement(self, stmt_addr) -> None:
        if stmt_addr in self.selected_ail_statements:
            self.selected_ail_statements.remove(stmt_addr)
            self.selected_ail_statements.am_event()
        self._update_published_view_state()

    def unselect_all_ail_statements(self) -> None:
        if self.selected_ail_statements:
            self.selected_ail_statements.clear()
            self.selected_ail_statements.am_event()
        self._update_published_view_state()

    def select_operand(
        self, ins_addr: int, operand_index: int, operand: OperandDescriptor, unique: bool = False
    ) -> None:
        """
        Mark an operand as selected.

        :param ins_addr:                Address of the instruction.
        :param operand_index:           Index of the operand.
        :param operand:   Data of the operand.
        :param unique:                 If this is a unique selection or not.
        :return:                            None
        """

        tpl = ins_addr, operand_index
        if tpl not in self.selected_operands:
            if unique:
                self.selected_operands.clear()
            self.selected_operands[tpl] = operand
            self.selected_operands.am_event()

    def unselect_operand(self, insn_addr, operand_idx) -> None:
        if (insn_addr, operand_idx) in self.selected_operands:
            self.selected_operands.pop((insn_addr, operand_idx))
            self.selected_operands.am_event()

    def unselect_all_operands(self) -> None:
        if self.selected_operands:
            self.selected_operands.clear()
            self.selected_operands.am_event()

    def select_label(self, label_addr) -> None:
        arch = self.disasm_view.instance.project.arch
        real_addr = get_real_address_if_arm(arch, label_addr)
        self.disasm_view.set_synchronized_cursor_address(real_addr)

        # only one label can be selected at a time
        # also, clear selection of instructions and operands
        self.unselect_all_instructions()
        self.unselect_all_operands()
        self.unselect_block_tree_node()

        self.selected_labels.clear()
        self.selected_labels.add(label_addr)
        self.selected_labels.am_event()

        self._update_published_view_state()

    def toggle_label_selection(self, addr: int) -> None:
        """
        Toggle the selection state of a label in the disassembly view.

        :param addr:    Address of the instruction to toggle.
        """

        if addr in self.selected_labels:
            self.unselect_label(addr)
        else:
            self.select_label(addr)

    def unselect_label(self, label_addr) -> None:
        if label_addr in self.selected_labels:
            self.selected_labels.remove(label_addr)
            self.selected_labels.am_event()
        self._update_published_view_state()

    def unselect_all_labels(self) -> None:
        self.selected_labels.clear()
        self.selected_labels.am_event()
        self._update_published_view_state()

    def toggle_instruction_selection(self, insn_addr, insn_pos=None, unique: bool = False) -> None:
        """
        Toggle the selection state of an instruction in the disassembly view.

        :param int insn_addr: Address of the instruction to toggle.
        :return:              None
        """

        if insn_addr in self.selected_insns:
            self.unselect_instruction(insn_addr)
        else:
            self.select_instruction(insn_addr, unique=unique, insn_pos=insn_pos)

    def toggle_operand_selection(self, insn_addr, operand_idx, operand, insn_pos=None, unique: bool = False) -> bool:
        """
        Toggle the selection state of an operand of an instruction in the disassembly view.

        :param int insn_addr:   Address of the instruction to toggle.
        :param int operand_idx: The operand to toggle.
        :param operand:         The operand instance.
        :return:                True if this operand is now selected, False otherwise.
        :rtype:                 bool
        """

        if (insn_addr, operand_idx) in self.selected_operands:
            self.unselect_operand(insn_addr, operand_idx)
            return False
        else:
            self.select_operand(insn_addr, operand_idx, operand, unique=unique)
            self.disasm_view.current_graph.show_instruction(insn_addr, insn_pos=insn_pos)
            return True

    def select_variable(self, unified_variable: SimVariable, unique: bool = True) -> None:
        self.unselect_all_labels()
        self.unselect_all_instructions()
        self.unselect_all_operands()

        if unique:
            self.selected_variables.clear()
        if unified_variable not in self.selected_variables:
            self.selected_variables.add(unified_variable)
            self.selected_variables.am_event()

    def unselect_variable(self, unified_variable: SimVariable) -> None:
        if unified_variable in self.selected_variables:
            self.selected_variables.remove(unified_variable)
            self.selected_variables.am_event()

    def toggle_variable_selection(self, unified_variable: SimVariable, unique: bool = True) -> bool:
        if len(self.selected_variables) > 1 and unique:
            # multiple variables are selected
            # clear existing selections and select this one
            self.select_variable(unified_variable, unique=True)
            return True

        if unified_variable in self.selected_variables:
            self.unselect_variable(unified_variable)
            return False
        else:
            self.select_variable(unified_variable, unique=unique)
            return True

    def clear_selection(self) -> None:
        self.selected_blocks.clear()
        self.selected_blocks.am_event()

        self.selected_insns.clear()
        self.selected_insns.am_event()

        self.selected_operands.clear()
        self.selected_operands.am_event()

        self.selected_variables.clear()
        self.selected_variables.am_event()

        self.selected_block_tree_node.am_obj = None
        self.selected_block_tree_node.am_event()

        self._update_published_view_state()

    def is_edge_hovered(self, src_addr, dst_addr):
        return self.hovered_edge.am_obj == (src_addr, dst_addr)

    def is_block_hovered(self, block_addr):
        return block_addr == self.hovered_block.am_obj

    def is_block_selected(self, block_addr) -> bool:
        return block_addr in self.selected_blocks

    def is_instruction_selected(self, ins_addr) -> bool:
        """
        Check if an instruction at @ins_addr is currently selected or not.

        :param int ins_addr:    Address of the instruction.
        :return:                True if it is selected, False otherwise.
        :rtype:                 bool
        """
        return ins_addr in self.selected_insns

    def is_operand_selected(self, ins_addr, operand_index) -> bool:
        """
        Check if an operand at @ins_addr and @operand_index is currently selected or not.

        :param int ins_addr:        Address of the instruction
        :param int operand_index:   Index of the operand.
        :return:                    bool
        """
        return (ins_addr, operand_index) in self.selected_operands

    def is_label_selected(self, label_addr) -> bool:
        return label_addr in self.selected_labels

    def is_variable_selected(self, unique_variable_ident: str) -> bool:
        return unique_variable_ident in self.selected_variables

    def select_block_tree_node(self, obj) -> None:
        """
        For QBlockCodeObj, we simply track selected state for now and handle
        matching in object handlers
        """

        self.unselect_all_instructions()
        self.unselect_all_labels()

        self.selected_block_tree_node.am_obj = obj
        self.selected_block_tree_node.am_event()

    def unselect_block_tree_node(self):
        self.selected_block_tree_node.am_obj = None
        self.selected_block_tree_node.am_event()

    def toggle_block_tree_node_selection(self, obj) -> None:
        """
        Toggle the selection state of a QBlockCodeObj in the disassembly view.

        :param obj:    The QBlockCodeObj to toggle.
        """

        if self.selected_block_tree_node.am_obj == obj:
            self.unselect_block_tree_node()
        else:
            self.select_block_tree_node(obj)

    def _update_published_view_state(self) -> None:
        self.disasm_view.published_view_state.cursors = self.selected_insns.union(self.selected_labels)
        self.disasm_view.notify_view_state_updated()



from __future__ import annotations

from typing import TYPE_CHECKING

from angrmanagement.data.analysis_options import AnalysisConfiguration

from .job import InstanceJob

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext


class CodeTaggingConfiguration(AnalysisConfiguration):
    """
    Configuration for Code Tagging.
    """

    def __init__(self, instance: Instance) -> None:
        super().__init__(instance)
        self.name = "code_tagging"
        self.display_name = "Tag Functions Based on Syntactic Features"
        self.description = "Add tags to functions based on syntactic features in assembly code and referenced strings."
        self.enabled = False


class CodeTaggingJob(InstanceJob):
    """
    Job for tagging functions.
    """

    def __init__(self, instance: Instance, on_finish=None) -> None:
        super().__init__("Code tagging", instance, on_finish=on_finish)

    def run(self, ctx: JobContext) -> None:
        func_count = len(self.instance.kb.functions)
        for i, func in enumerate(self.instance.kb.functions.values()):
            if func.is_alignment:
                continue
            ct = self.instance.project.analyses.CodeTagging(func)
            func.tags = tuple(ct.tags)

            percentage = i / func_count * 100
            ctx.set_progress(percentage)

    def __repr__(self) -> str:
        return "CodeTaggingJob"



from __future__ import annotations

import logging
import os
import sys
from string import Template

from PySide6.QtGui import QPalette
from PySide6.QtWidgets import QApplication

from angrmanagement.config import Conf
from angrmanagement.consts import RES_LOCATION, THEME_LOCATION
from angrmanagement.data.object_container import ObjectContainer
from angrmanagement.logic import GlobalInfo
from angrmanagement.ui.icons import update_icon_colors
from angrmanagement.ui.widgets.qccode_highlighter import reset_formats

log = logging.getLogger(__name__)


#
# FIXME: Need to propagate CSS updates to more objects.
#


class CSS:
    """
    Main stylesheet re-load logic.
    """

    global_css = ObjectContainer("", "Global CSS")

    @staticmethod
    def rebuild() -> None:
        base_css_path = os.path.join(THEME_LOCATION, "base.css")
        try:
            with open(base_css_path, encoding="utf-8") as f:
                css = f.read()
        except Exception:  # pylint: disable=broad-except
            log.warning("Failed to load base theme at %s", base_css_path)
            css = ""

        theme_path = os.path.join(THEME_LOCATION, Conf.base_theme_name)
        css_path = os.path.join(theme_path, "theme.css")
        if os.path.exists(css_path):
            try:
                with open(css_path, encoding="utf-8") as f:
                    css += "\n" + f.read()
            except Exception:  # pylint: disable=broad-except
                log.warning("Failed to load theme CSS at %s", css_path)

        theme_resources_path = RES_LOCATION
        if sys.platform == "win32":
            theme_resources_path = theme_resources_path.replace("\\", "/")
            theme_path = theme_path.replace("\\", "/")
        CSS.global_css.am_obj = Template(css).safe_substitute(resources=theme_resources_path, theme=theme_path)
        CSS.global_css.am_event()


def refresh_theme() -> None:
    app = QApplication.instance()

    # determine the default application style according to the OS
    if sys.platform == "win32" and Conf.base_theme_name == "Light":
        app_style = "windowsvista"
    else:
        app_style = "Fusion"

    if app_style:
        app.setStyle(app_style)

    palette = QPalette()
    palette.setColor(QPalette.ColorRole.Window, Conf.palette_window)
    palette.setColor(QPalette.ColorRole.WindowText, Conf.palette_windowtext)
    palette.setColor(QPalette.ColorRole.Base, Conf.palette_base)
    palette.setColor(QPalette.ColorRole.AlternateBase, Conf.palette_alternatebase)
    palette.setColor(QPalette.ColorRole.ToolTipBase, Conf.palette_tooltipbase)
    palette.setColor(QPalette.ColorRole.ToolTipText, Conf.palette_tooltiptext)
    palette.setColor(QPalette.ColorRole.PlaceholderText, Conf.palette_placeholdertext)
    palette.setColor(QPalette.ColorRole.Text, Conf.palette_text)
    palette.setColor(QPalette.ColorRole.Button, Conf.palette_button)
    palette.setColor(QPalette.ColorRole.ButtonText, Conf.palette_buttontext)
    palette.setColor(QPalette.ColorRole.BrightText, Conf.palette_brighttext)
    palette.setColor(QPalette.ColorRole.Highlight, Conf.palette_highlight)
    palette.setColor(QPalette.ColorRole.HighlightedText, Conf.palette_highlightedtext)
    palette.setColor(QPalette.ColorRole.Light, Conf.palette_light)
    palette.setColor(QPalette.ColorRole.Midlight, Conf.palette_midlight)
    palette.setColor(QPalette.ColorRole.Dark, Conf.palette_dark)
    palette.setColor(QPalette.ColorRole.Mid, Conf.palette_mid)
    palette.setColor(QPalette.ColorRole.Shadow, Conf.palette_shadow)
    palette.setColor(QPalette.ColorRole.Link, Conf.palette_link)
    palette.setColor(QPalette.ColorRole.LinkVisited, Conf.palette_linkvisited)
    palette.setColor(QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text, Conf.palette_disabled_text)
    palette.setColor(QPalette.ColorGroup.Disabled, QPalette.ColorRole.ButtonText, Conf.palette_disabled_buttontext)
    palette.setColor(QPalette.ColorGroup.Disabled, QPalette.ColorRole.WindowText, Conf.palette_disabled_windowtext)
    update_icon_colors(palette)
    app.setPalette(palette)
    CSS.rebuild()
    app.setStyleSheet(CSS.global_css.am_obj)

    reset_formats()

    if GlobalInfo.main_window is not None:
        for codeview in GlobalInfo.main_window.workspace.view_manager.views_by_category["pseudocode"]:
            codeview.set_codeedit_palette(palette)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from PySide6.QtCore import QThread

    from angrmanagement.data.library_docs import LibraryDocs
    from angrmanagement.plugins.plugin_manager import PluginManager
    from angrmanagement.ui.main_window import MainWindow, QExecuteCodeEventReceiver


class GlobalInfo:
    """
    Global data.
    """

    gui_thread: QThread | None = None
    main_window: MainWindow | None = None
    event_receiver: QExecuteCodeEventReceiver
    daemon_inst = None
    daemon_conn = None
    client_id: str | None = None
    headless_plugin_manager: PluginManager | None = None
    library_docs: LibraryDocs = None
    autoreload = False



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtGui import Qt
from PySide6.QtWidgets import QDialog, QGridLayout, QLabel, QTextEdit, QVBoxLayout

from angrmanagement.config import Conf

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance


class FuncDocDialog(QDialog):
    """
    Implements the FuncDoc dialog.
    """

    def __init__(
        self, instance: Instance, addr: int | None = None, name: str = "", doc_tuple=None, parent=None
    ) -> None:
        super().__init__(parent)

        # initialization
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowType.WindowContextHelpButtonHint)

        self.instance = instance
        self._addr = addr
        self._name = name
        self._doc = doc_tuple[0].strip()
        self._url = doc_tuple[1].strip()
        self._ftype = doc_tuple[2].strip()
        self.setWindowTitle("Function Documentation")
        self.main_layout = QVBoxLayout()
        self._init_widgets()
        self.setLayout(self.main_layout)

    def _init_widgets(self) -> None:
        layout = QGridLayout()

        # validation_failures = set()
        addr = hex(self._addr)
        address_label = QLabel(self)
        address_label.setText(f"Function at address {addr}: {self._name}")

        layout.addWidget(address_label)

        type_label = QLabel(self)
        type_label.setText(f"Type: {self._ftype}")

        layout.addWidget(type_label)

        text_edit = QTextEdit(self)
        text_edit.setMinimumWidth(800)
        text_edit.setMinimumHeight(450)
        text_edit.setFont(Conf.disasm_font)
        text_edit.setText(self._doc)

        url_label = QLabel(self)
        hyperlink = f'<a href="{self._url}">{self._url}</a>'
        url_label.setText(hyperlink)
        url_label.setOpenExternalLinks(True)

        layout.addWidget(text_edit)
        layout.addWidget(url_label)

        self.main_layout.addLayout(layout)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .job import InstanceJob

if TYPE_CHECKING:
    from angr.knowledge_plugins.functions.function import Function

    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext

log = logging.getLogger(__name__)


class LLMRefineJob(InstanceJob):
    """
    Job that runs LLM-based refinement on a decompiled function.
    Supports different modes: suggest variable names, function name, variable types, or all.
    """

    SUGGEST_VARIABLE_NAMES = "variable_names"
    SUGGEST_FUNCTION_NAME = "function_name"
    SUGGEST_VARIABLE_TYPES = "variable_types"
    SUMMARIZE = "summarize"
    REFINE_ALL = "all"

    def __init__(
        self,
        instance: Instance,
        function: Function,
        mode: str = "all",
        on_finish=None,
        blocking: bool = False,
    ) -> None:
        super().__init__("LLM Refine", instance, on_finish=on_finish, blocking=blocking)
        self.function = function
        self.mode = mode

    def run(self, ctx: JobContext) -> None:
        project = self.instance.project.am_obj
        if project is None:
            log.warning("LLMRefineJob: no project available")
            return

        llm_client = project.llm_client
        if llm_client is None:
            log.warning("LLMRefineJob: no LLM client configured")
            return

        # get cached decompilation
        key = self.function.addr, "pseudocode"
        if key not in self.instance.kb.decompilations:
            log.warning("LLMRefineJob: no cached decompilation for %s", self.function.name)
            return
        dec_cache = self.instance.kb.decompilations[key]
        if dec_cache is None or dec_cache.codegen is None:
            log.warning("LLMRefineJob: no cached decompilation for %s", self.function.name)
            return

        ctx.set_progress(0, "Setting up decompiler")

        # create a Decompiler instance without decompiling, then attach existing codegen
        dec = project.analyses.Decompiler(
            self.function,
            variable_kb=self.instance.pseudocode_variable_kb,
            decompile=False,
        )
        dec.codegen = dec_cache.codegen
        dec.cache = dec_cache

        changed = False
        if self.mode in (self.REFINE_ALL, self.SUGGEST_VARIABLE_NAMES):
            ctx.set_progress(10, "Suggesting variable names...")
            changed |= dec.llm_suggest_variable_names(llm_client=llm_client)

        if self.mode in (self.REFINE_ALL, self.SUGGEST_FUNCTION_NAME):
            ctx.set_progress(40, "Suggesting function name...")
            changed |= dec.llm_suggest_function_name(llm_client=llm_client)

        if self.mode in (self.REFINE_ALL, self.SUGGEST_VARIABLE_TYPES):
            ctx.set_progress(70, "Suggesting variable types...")
            changed |= dec.llm_suggest_variable_types(llm_client=llm_client)

        if changed:
            ctx.set_progress(90, "Regenerating text...")
            dec.codegen.regenerate_text()
            dec_cache.codegen = dec.codegen

        if self.mode in (self.REFINE_ALL, self.SUMMARIZE):
            ctx.set_progress(95, "Summarizing function...")
            dec.llm_summarize_function(llm_client=llm_client)

        ctx.set_progress(100, "Done")



from __future__ import annotations

from .server import daemon_conn, daemon_exists, run_daemon_process, start_daemon
from .url_handler import handle_url


def monkeypatch_rpyc() -> None:
    # The AsyncResult in rpyc has a bug that causes a race condition when clients are cascaded. this monkeypatch fixes
    # the bug.
    from rpyc.core.async_ import AsyncResult, AsyncResultTimeout

    def patched_wait(self):
        while not self._is_ready and not self._ttl.expired():
            self._conn.serve(0.01)
        if not self._is_ready:
            raise AsyncResultTimeout("result expired")

    AsyncResult.wait = patched_wait


monkeypatch_rpyc()

__all__ = [
    "daemon_conn",
    "daemon_exists",
    "run_daemon_process",
    "start_daemon",
    "handle_url",
]



"""
Goto palette dialog for navigation.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import Qt
from PySide6.QtGui import QColor

from angrmanagement.config import Conf

from .palette import PaletteDialog, PaletteModel

if TYPE_CHECKING:
    from angr.knowledge_plugins.functions import Function

    from angrmanagement.ui.workspace import Workspace


class GotoPaletteModel(PaletteModel):
    """
    Data provider for goto palette.
    """

    def get_items(self) -> list[Function]:
        items = []

        instance = self.workspace.main_instance
        if instance and not instance.project.am_none:
            project = instance.project.am_obj
            if project is not None:
                items.extend([func for _, func in project.kb.functions.items()])

        return items

    def get_icon_color_and_text_for_item(self, item: Function) -> tuple[QColor | None, str]:
        if item.is_syscall:
            color = Conf.function_table_syscall_color
        elif item.is_plt:
            color = Conf.function_table_plt_color
        elif item.is_simprocedure:
            color = Conf.function_table_simprocedure_color
        elif item.is_alignment:
            color = Conf.function_table_alignment_color
        else:
            color = QColor(Qt.GlobalColor.gray)
        return (color, "f")

    def get_caption_for_item(self, item: Function) -> str:
        return item.name

    def get_annotation_for_item(self, item: Function) -> str:
        return f"{item.addr:x}"


class GotoPaletteDialog(PaletteDialog):
    """
    Dialog for selecting navigation targets.
    """

    def __init__(self, workspace: Workspace, parent=None) -> None:
        super().__init__(GotoPaletteModel(workspace), parent=parent)
        self.setWindowTitle("Goto Anything")



from __future__ import annotations

from typing import TYPE_CHECKING

import networkx
from angr import SIM_PROCEDURES
from angr.code_location import ExternalCodeLocation
from PySide6.QtCore import QSize
from PySide6.QtWidgets import QHBoxLayout

from angrmanagement.ui.widgets.qdep_graph import QDependencyGraph
from angrmanagement.ui.widgets.qdepgraph_block import QDepGraphBlock

from .view import InstanceView

if TYPE_CHECKING:
    from angr.knowledge_plugins.key_definitions.atoms import Atom
    from angr.knowledge_plugins.key_definitions.definition import Definition

    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.workspace import Workspace


class DependencyView(InstanceView):
    """
    Creates view for dependency analysis.
    """

    def __init__(self, workspace: Workspace, default_docking_position: str, instance: Instance) -> None:
        super().__init__("dependencies", workspace, default_docking_position, instance)

        self.base_caption = "Dependencies"

        # UI widgets
        self._graph_widget: QDependencyGraph = None

        # data
        self.sink_atom: Atom | None = None
        self.sink_ins_addr: int | None = None
        self.closures: dict[Definition, networkx.DiGraph] | None = None
        self._graph: networkx.DiGraph | None = None
        self.hovered_block: QDepGraphBlock | None = None

        self._init_widgets()
        self._register_events()

    def hover_enter_block(self, block: QDepGraphBlock) -> None:
        self.hovered_block = block
        if self._graph_widget is not None:
            self._graph_widget.on_block_hovered(block)
        self.redraw_graph()

    def hover_leave_block(self) -> None:
        self.hovered_block = None
        self.redraw_graph()

    def on_screen_changed(self) -> None:
        if self._graph_widget is not None:
            self._graph_widget.refresh()

    def reload(self) -> None:
        if self._graph_widget is None:
            return
        # re-generate the graph
        if not self.closures:
            self._graph = None
            self._graph_widget.graph = None
            self._graph_widget.request_relayout()
            return

        self._graph = self._create_ui_graph()
        self._graph_widget.graph = self._graph

    def redraw_graph(self) -> None:
        if self._graph_widget is not None:
            self._graph_widget.viewport().update()

    def sizeHint(self):
        return QSize(400, 800)

    def _init_widgets(self) -> None:
        self._graph_widget = QDependencyGraph(self.instance, self)

        hlayout = QHBoxLayout()
        hlayout.addWidget(self._graph_widget)
        hlayout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(hlayout)

    def _register_events(self) -> None:
        self.workspace.current_screen.am_subscribe(self.on_screen_changed)

    def _convert_node(self, node: Definition, converted: dict[Definition, QDepGraphBlock]) -> QDepGraphBlock | None:
        if node in converted:
            return converted[node]

        # skip external
        if isinstance(node.codeloc, ExternalCodeLocation):
            return None

        if self.instance.project.is_hooked(node.codeloc.block_addr):
            hook = self.instance.project.hooked_by(node.codeloc.block_addr)
            if isinstance(
                hook,
                SIM_PROCEDURES["stubs"]["UnresolvableJumpTarget"] | SIM_PROCEDURES["stubs"]["UnresolvableCallTarget"],
            ):
                return None

        new_node = QDepGraphBlock(False, self, definition=node, addr=node.codeloc.ins_addr)
        converted[node] = new_node
        return new_node

    # def _is_edge_in_graph(self):
    def _create_ui_graph(self) -> networkx.DiGraph:
        g = networkx.DiGraph()
        source_node = QDepGraphBlock(False, self, atom=self.sink_atom, addr=self.sink_ins_addr)
        g.add_node(source_node)
        all_graphs = networkx.compose_all(self.closures.values())
        converted = {}

        for node_ in all_graphs.nodes:
            node = self._convert_node(node_, converted)
            if node is not None:
                g.add_node(node)
            # this is a hack - we only want our sink as the only root of the dependency tree
            # TODO: Figure out why
            if all_graphs.out_degree[node_] == 0:
                g.add_edge(node, source_node)

        for src_, dst_ in all_graphs.edges:
            src = self._convert_node(src_, converted)
            dst = self._convert_node(dst_, converted)
            if src is not None and dst is not None:
                g.add_edge(src, dst)

        return g



# pylint:disable=missing-class-docstring,wrong-import-order
from __future__ import annotations

import os
import sys
import unittest

import angr
from common import AngrManagementTestCase, test_location

from angrmanagement.ui.views import DisassemblyView


class TestUndefineRedefineCodeRegions(AngrManagementTestCase):
    def test_undefine_redefine_code_region_in_disasm_view(self):
        main = self.main
        binpath = os.path.join(test_location, "x86_64", "fauxware")
        main.workspace.main_instance.project.am_obj = angr.Project(binpath, auto_load_libs=False)
        main.workspace.main_instance.project.am_event()
        main.workspace.job_manager.join_all_jobs()

        cfg = main.workspace.main_instance.cfg
        func = main.workspace.main_instance.project.kb.functions["main"]
        assert func is not None

        insn_addr = 0x40073E
        insn_node = cfg.get_any_node(insn_addr)
        assert insn_node is not None
        assert insn_node.function_address == 0x40071D

        # load the disassembly view
        disasm_view = main.workspace._get_or_create_view("disassembly", DisassemblyView)
        disasm_view.display_disasm_graph()
        disasm_view.display_function(func)

        # select the instruction right after call to puts
        disasm_view.infodock.select_instruction(0x40073E)
        assert len(disasm_view.infodock.selected_insns) == 1
        assert disasm_view.infodock.selected_insns == {0x40073E}

        # undefine the instruction
        disasm_view.undefine_code()

        main.workspace.job_manager.join_all_jobs()

        # Node at 0x40073E should be removed from the CFG
        assert cfg.get_any_node(0x40073E) is None

        main.workspace.job_manager.join_all_jobs()

        # select the instruction right after call to puts and redefine it as code
        disasm_view.infodock.select_label(0x40073E)
        assert disasm_view.infodock.selected_labels == {0x40073E}
        disasm_view.define_code()
        main.workspace.job_manager.join_all_jobs()

        # Redefined node should be part of main function
        insn_node = cfg.get_any_node(insn_addr)
        assert insn_node is not None
        assert insn_node.function_address == 0x40071D


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



"""
Test cases for generic Palette framework (base classes).
"""
# pylint: disable=no-self-use

from __future__ import annotations

import unittest
from unittest.mock import MagicMock, patch

from common import AngrManagementTestCase
from PySide6.QtCore import Qt
from PySide6.QtGui import QKeyEvent
from PySide6.QtWidgets import QApplication

from angrmanagement.logic.commands import BasicCommand
from angrmanagement.logic.commands.command_manager import CommandManager
from angrmanagement.ui.dialogs.command_palette import CommandPaletteModel
from angrmanagement.ui.dialogs.palette import (
    PaletteDialog,
    PaletteItemDelegate,
    PaletteModel,
)


class SimplePaletteModel(PaletteModel):
    """Simple test model for testing base palette functionality."""

    def __init__(self, workspace, items=None):
        self._test_items = items or []
        super().__init__(workspace)

    def get_items(self):
        return self._test_items

    def get_caption_for_item(self, item):
        return item if isinstance(item, str) else str(item)


class TestPaletteModelBase(unittest.TestCase):
    """Test PaletteModel base class functionality."""

    def setUp(self):
        self.mock_workspace = MagicMock()

    def test_base_model_get_items(self):
        """Test that base PaletteModel.get_items() returns empty list."""
        model = PaletteModel(self.mock_workspace)
        items = model.get_items()
        assert not items

    def test_base_model_get_caption_for_item(self):
        """Test that base PaletteModel.get_caption_for_item() returns empty string."""
        model = PaletteModel(self.mock_workspace)
        caption = model.get_caption_for_item("test_item")
        assert caption == ""

    def test_base_model_get_subcaption_for_item(self):
        """Test that base PaletteModel.get_subcaption_for_item() returns None."""
        model = PaletteModel(self.mock_workspace)
        assert model.get_subcaption_for_item("test_item") is None

    def test_base_model_get_annotation_for_item(self):
        """Test that base PaletteModel.get_annotation_for_item() returns None."""
        model = PaletteModel(self.mock_workspace)
        assert model.get_annotation_for_item("test_item") is None

    def test_base_model_get_icon_color_and_text_for_item(self):
        """Test that base PaletteModel.get_icon_color_and_text_for_item() returns None and empty string."""
        model = PaletteModel(self.mock_workspace)
        color, text = model.get_icon_color_and_text_for_item("test_item")
        assert color is None
        assert text == ""

    def test_row_count_with_no_filter(self):
        """Test that rowCount returns all items when no filter is applied."""
        model = SimplePaletteModel(self.mock_workspace, ["item1", "item2", "item3"])
        assert model.rowCount() == 3

    def test_column_count(self):
        """Test that columnCount returns 1."""
        model = SimplePaletteModel(self.mock_workspace)
        assert model.columnCount() == 1

    def test_set_filter_text_empty(self):
        """Test that set_filter_text with empty string shows all items."""
        model = SimplePaletteModel(self.mock_workspace, ["Test Item", "Other Item"])
        original_count = model.rowCount()

        model.set_filter_text("Test")
        model.set_filter_text("")

        assert model.rowCount() == original_count

    def test_set_filter_text_filters_by_caption(self):
        """Test that set_filter_text filters items by matching caption."""
        model = SimplePaletteModel(self.mock_workspace, ["Test Item", "Other Item"])
        model.set_filter_text("Test")

        filtered_items = [model.data(model.index(i, 0)) for i in range(model.rowCount())]
        captions = [model.get_caption_for_item(item) for item in filtered_items if item is not None]

        assert any("Test" in caption for caption in captions)

    def test_set_filter_text_fuzzy_matches_with_typos(self):
        """Test that fuzzy matching finds items even with typos."""
        model = SimplePaletteModel(self.mock_workspace, ["Test Command", "Other Command"])
        # "Tst Commnd" should fuzzy match "Test Command"
        model.set_filter_text("Tst Commnd")

        filtered_items = [model.data(model.index(i, 0)) for i in range(model.rowCount())]
        captions = [model.get_caption_for_item(item) for item in filtered_items if item is not None]

        # Should find Test Command despite typos
        assert any("Test Command" in caption for caption in captions)

    def test_set_filter_text_fuzzy_matches_partial(self):
        """Test that fuzzy matching finds items with partial text."""
        model = SimplePaletteModel(self.mock_workspace, ["Open File", "Save File", "Close Window"])
        # "File" should match both "Open File" and "Save File"
        model.set_filter_text("File")

        filtered_items = [model.data(model.index(i, 0)) for i in range(model.rowCount())]
        captions = [model.get_caption_for_item(item) for item in filtered_items if item is not None]

        # Should find both file-related commands
        assert any("Open File" in caption for caption in captions)
        assert any("Save File" in caption for caption in captions)
        # Should return some results (fuzzy matching is working)
        assert len(captions) > 0

    def test_set_filter_text_fuzzy_matches_abbreviations(self):
        """Test that fuzzy matching finds items using abbreviations."""
        model = SimplePaletteModel(self.mock_workspace, ["Test Command", "Other Item"])
        # "TC" should fuzzy match "Test Command" (first letters)
        model.set_filter_text("TC")

        filtered_items = [model.data(model.index(i, 0)) for i in range(model.rowCount())]
        captions = [model.get_caption_for_item(item) for item in filtered_items if item is not None]

        # Should find Test Command via abbreviation
        assert any("Test Command" in caption for caption in captions)

    def test_data_returns_item(self):
        """Test that data returns the correct item."""
        model = SimplePaletteModel(self.mock_workspace, ["Test Item"])

        index = model.index(0, 0)
        data = model.data(index)
        assert data == "Test Item"

    def test_data_returns_none_for_invalid_index(self):
        """Test that data returns None for invalid index."""
        model = SimplePaletteModel(self.mock_workspace)
        index = model.index(-1, 0)
        assert model.data(index) is None

    def test_parent_returns_empty_index(self):
        """Test that parent returns empty QModelIndex."""
        model = SimplePaletteModel(self.mock_workspace, ["Test Item"])
        index = model.index(0, 0)
        parent_index = model.parent(index)
        assert not parent_index.isValid()


class TestPaletteDialog(AngrManagementTestCase):
    """Test PaletteDialog base functionality."""

    def setUp(self):
        super().setUp()
        mock_workspace = MagicMock()
        mock_workspace.command_manager = CommandManager()
        self.model = CommandPaletteModel(mock_workspace)
        self.dialog = PaletteDialog(self.model, parent=self.main)

    def tearDown(self):
        if hasattr(self, "dialog"):
            self.dialog.close()
            del self.dialog
        super().tearDown()

    def test_dialog_initialization(self):
        """Test that PaletteDialog initializes correctly."""
        assert self.dialog._model == self.model
        assert self.dialog.selected_item is None
        assert self.dialog.windowTitle() == "Palette"

    def test_dialog_has_query_input(self):
        """Test that dialog has query input field."""
        assert self.dialog._query is not None
        assert hasattr(self.dialog._query, "text")

    def test_dialog_has_list_view(self):
        """Test that dialog has list view."""
        assert self.dialog._view is not None
        assert self.dialog._view.model() == self.model

    def test_query_text_changed_updates_filter(self):
        """Test that changing query text updates model filter."""
        with patch.object(self.model, "set_filter_text") as mock_set_filter:
            self.dialog._query.setText("test")
            QApplication.processEvents()
            mock_set_filter.assert_called()

    def test_up_key_navigates_list(self):
        """Test that Up key is passed to list view."""
        with patch.object(self.dialog._view, "keyPressEvent") as mock_key_press:
            key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_Up, Qt.KeyboardModifier.NoModifier)
            self.dialog.keyPressEvent(key_event)
            mock_key_press.assert_called_once()

    def test_down_key_navigates_list(self):
        """Test that Down key is passed to list view."""
        with patch.object(self.dialog._view, "keyPressEvent") as mock_key_press:
            key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_Down, Qt.KeyboardModifier.NoModifier)
            self.dialog.keyPressEvent(key_event)
            mock_key_press.assert_called_once()

    def test_enter_key_accepts_dialog(self):
        """Test that Enter key accepts the dialog."""
        cmd = BasicCommand("test_cmd", "Test Command", MagicMock())
        self.model.workspace.command_manager.register_command(cmd)

        with patch.object(self.dialog, "accept") as mock_accept:
            key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_Enter, Qt.KeyboardModifier.NoModifier)
            self.dialog.keyPressEvent(key_event)
            mock_accept.assert_called_once()

    def test_return_key_accepts_dialog(self):
        """Test that Return key accepts the dialog."""
        cmd = BasicCommand("test_cmd", "Test Command", MagicMock())
        self.model.workspace.command_manager.register_command(cmd)

        with patch.object(self.dialog, "accept") as mock_accept:
            key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_Return, Qt.KeyboardModifier.NoModifier)
            self.dialog.keyPressEvent(key_event)
            mock_accept.assert_called_once()

    def test_accept_sets_selected_item(self):
        """Test that accept sets the selected_item."""
        cmd = BasicCommand("test_cmd", "Test Command", MagicMock())
        mock_workspace = MagicMock()
        mock_workspace.command_manager = CommandManager()
        mock_workspace.command_manager.register_command(cmd)

        model = CommandPaletteModel(mock_workspace)
        dialog = PaletteDialog(model, parent=self.main)

        index = model.index(0, 0)
        dialog._view.setCurrentIndex(index)

        with patch("PySide6.QtWidgets.QDialog.accept"):
            dialog.accept()

        assert dialog.selected_item is not None
        dialog.close()

    def test_get_selected_returns_none_when_nothing_selected(self):
        """Test that _get_selected returns None when nothing is selected."""
        # Clear any auto-selection
        self.dialog._view.clearSelection()
        QApplication.processEvents()

        result = self.dialog._get_selected()
        assert result is None

    def test_other_keys_passed_to_parent(self):
        """Test that other keys are passed to parent keyPressEvent."""
        with patch("PySide6.QtWidgets.QDialog.keyPressEvent") as mock_parent_key_press:
            key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_A, Qt.KeyboardModifier.NoModifier)
            self.dialog.keyPressEvent(key_event)
            mock_parent_key_press.assert_called_once()


class TestPaletteItemDelegate(AngrManagementTestCase):
    """Test PaletteItemDelegate functionality."""

    def test_delegate_initialization_with_icons(self):
        """Test that PaletteItemDelegate initializes with icons enabled."""
        delegate = PaletteItemDelegate(display_icons=True)
        assert delegate._display_icons is True

    def test_delegate_initialization_without_icons(self):
        """Test that PaletteItemDelegate initializes with icons disabled."""
        delegate = PaletteItemDelegate(display_icons=False)
        assert delegate._display_icons is False

    def test_delegate_get_text_document(self):
        """Test that _get_text_document returns QTextDocument."""
        mock_workspace = MagicMock()
        cmd = BasicCommand("test_cmd", "Test Command", MagicMock())

        mock_command_manager = MagicMock()
        mock_command_manager.get_commands.return_value = [cmd]
        mock_workspace.command_manager = mock_command_manager

        model = CommandPaletteModel(mock_workspace)
        delegate = PaletteItemDelegate()

        index = model.index(0, 0)
        text_doc = delegate._get_text_document(index)

        assert text_doc is not None
        assert hasattr(text_doc, "toHtml")

    def test_delegate_get_text_document_with_filter(self):
        """Test that _get_text_document bolds matching text when filter is applied."""
        mock_workspace = MagicMock()
        cmd = BasicCommand("test_cmd", "Test Command", MagicMock())

        mock_command_manager = MagicMock()
        mock_command_manager.get_commands.return_value = [cmd]
        mock_workspace.command_manager = mock_command_manager

        model = CommandPaletteModel(mock_workspace)
        model.set_filter_text("Test")
        delegate = PaletteItemDelegate()

        index = model.index(0, 0)
        text_doc = delegate._get_text_document(index)

        html = text_doc.toHtml()
        # Qt converts <b> to font-weight style in HTML
        assert "font-weight:700" in html or "<b>" in html

    def test_delegate_get_text_document_with_subcaption(self):
        """Test that _get_text_document includes subcaption when provided."""

        class ModelWithSubcaption(PaletteModel):
            """Test model with subcaption."""

            def get_items(self):
                return ["item1"]

            def get_caption_for_item(self, item):  # noqa: ARG002
                return "Main Caption"

            def get_subcaption_for_item(self, item):  # noqa: ARG002
                return "Subcaption Text"

        mock_workspace = MagicMock()
        model = ModelWithSubcaption(mock_workspace)
        delegate = PaletteItemDelegate()

        index = model.index(0, 0)
        text_doc = delegate._get_text_document(index)

        html = text_doc.toHtml()
        assert "Subcaption Text" in html
        assert "<sub>" in html or "Subcaption Text" in html

    def test_dialog_renders_items_with_annotations(self):
        """Test that dialog renders items with annotations without error."""

        class ModelWithAnnotation(PaletteModel):
            """Test model with annotation."""

            def get_items(self):
                return ["item1", "item2"]

            def get_caption_for_item(self, item):  # noqa: ARG002
                return "Test Item"

            def get_annotation_for_item(self, item):  # noqa: ARG002
                return "0x1000"

        mock_workspace = MagicMock()
        model = ModelWithAnnotation(mock_workspace)
        dialog = PaletteDialog(model, parent=self.main)

        # Select an item to trigger selected state rendering
        dialog._view.setCurrentIndex(model.index(0, 0))

        # Trigger rendering (exercises paint method for all items)
        dialog.show()
        QApplication.processEvents()

        # Force Qt to actually paint all items
        dialog._view.viewport().repaint()
        QApplication.processEvents()

        dialog.close()

    def test_dialog_renders_items_with_icon_variations(self):
        """Test that dialog renders items with different icon configurations."""

        class ModelWithIcons(PaletteModel):
            """Test model with various icon configurations."""

            def get_items(self):
                return ["color_and_text", "color_only", "text_only", "neither"]

            def get_caption_for_item(self, item):  # noqa: ARG002
                return "Test Item"

            def get_icon_color_and_text_for_item(self, item):  # noqa: ARG002  # type: ignore[override]
                if item == "color_and_text":
                    return ("#ff0000", "T")
                if item == "color_only":
                    return ("#00ff00", "")
                if item == "text_only":
                    return (None, "X")
                return (None, "")

        mock_workspace = MagicMock()
        model = ModelWithIcons(mock_workspace)
        dialog = PaletteDialog(model, parent=self.main)

        # Set delegate with icons enabled
        delegate = PaletteItemDelegate(display_icons=True)
        dialog._view.setItemDelegate(delegate)

        # Trigger rendering (exercises all icon variations)
        dialog.show()
        QApplication.processEvents()

        # Force Qt to actually paint all items
        dialog._view.viewport().repaint()
        QApplication.processEvents()

        dialog.close()

    def test_dialog_renders_items_without_icons(self):
        """Test that dialog renders items with icons disabled."""

        class SimpleModel(PaletteModel):
            """Test model with simple items."""

            def get_items(self):
                return ["item1", "item2"]

            def get_caption_for_item(self, item):  # noqa: ARG002
                return "Test Item"

        mock_workspace = MagicMock()
        model = SimpleModel(mock_workspace)
        dialog = PaletteDialog(model, parent=self.main)

        # Set delegate with icons disabled
        delegate = PaletteItemDelegate(display_icons=False)
        dialog._view.setItemDelegate(delegate)

        # Trigger rendering (exercises paint without icons)
        dialog.show()
        QApplication.processEvents()

        # Force Qt to actually paint all items
        dialog._view.viewport().repaint()
        QApplication.processEvents()

        dialog.close()



from __future__ import annotations

from .debugger import Debugger, DebuggerListManager, DebuggerManager, DebuggerWatcher

__all__ = [
    "Debugger",
    "DebuggerListManager",
    "DebuggerManager",
    "DebuggerWatcher",
]



# pylint:disable=missing-class-docstring,wrong-import-order
from __future__ import annotations

import os
import sys
import unittest

import angr
from common import AngrManagementTestCase, test_location

from angrmanagement.ui.views import DisassemblyView
from angrmanagement.ui.widgets import DisassemblyLevel


class TestQLinearViewer(AngrManagementTestCase):
    def test_disassembly_level_change(self):
        main = self.main
        binpath = os.path.join(test_location, "x86_64", "fauxware")
        main.workspace.main_instance.project.am_obj = angr.Project(binpath, auto_load_libs=False)
        main.workspace.main_instance.project.am_event()
        main.workspace.job_manager.join_all_jobs()

        func = main.workspace.main_instance.project.kb.functions["main"]
        assert func is not None

        # load the disassembly view
        disasm_view = main.workspace._get_or_create_view("disassembly", DisassemblyView)
        disasm_view.display_linear_viewer()
        disasm_view.display_function(func)

        # make sure function block is in cache
        assert func.addr in disasm_view.linear_viewer.objects
        assert disasm_view.linear_viewer.objects[func.addr] is not None
        assert disasm_view.linear_viewer.objects[func.addr].disassembly_level is DisassemblyLevel.MachineCode

        # change disassembly level to LifterIR
        disasm_view.set_disassembly_level_lifter_ir()
        main.workspace.job_manager.join_all_jobs()
        disasm_view.display_function(func)

        # the block should be reloaded and the disassembly level should be updated
        assert func.addr in disasm_view.linear_viewer.objects
        assert disasm_view.linear_viewer.objects[func.addr] is not None
        assert disasm_view.linear_viewer.objects[func.addr].disassembly_level is DisassemblyLevel.LifterIR

    def test_cfb_update(self):
        main = self.main
        binpath = os.path.join(test_location, "x86_64", "fauxware")
        main.workspace.main_instance.project.am_obj = angr.Project(binpath, auto_load_libs=False)
        main.workspace.main_instance.project.am_event()
        main.workspace.job_manager.join_all_jobs()

        func = main.workspace.main_instance.project.kb.functions["main"]
        assert func is not None

        # load the disassembly view
        disasm_view = main.workspace._get_or_create_view("disassembly", DisassemblyView)
        disasm_view.display_linear_viewer()
        disasm_view.display_function(func)

        # make sure function block is in cache
        assert func.addr in disasm_view.linear_viewer.objects
        assert disasm_view.linear_viewer.objects[func.addr] is not None
        assert len(disasm_view.linear_viewer.objects[func.addr].addr_to_insns) == 9

        # select the third instruction of the function
        disasm_view.infodock.select_instruction(func.addr + 4)
        assert len(disasm_view.infodock.selected_insns) == 1
        assert disasm_view.infodock.selected_insns == {func.addr + 4}

        # undefine third instruction to trigger cfb update
        disasm_view.undefine_code()
        main.workspace.job_manager.join_all_jobs()

        disasm_view.display_function(func)

        # the block should be updated to reflect removed instructions
        assert func.addr in disasm_view.linear_viewer.objects
        assert disasm_view.linear_viewer.objects[func.addr] is not None
        assert len(disasm_view.linear_viewer.objects[func.addr].addr_to_insns) == 2


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



from __future__ import annotations

from typing import TYPE_CHECKING

from angr.ailment.expression import BinaryOp, Load, Op, UnaryOp
from angr.ailment.statement import Assignment, Store
from angr.analyses.decompiler.optimization_passes.expr_op_swapper import OpDescriptor
from angr.analyses.decompiler.structured_codegen.c import (
    CBinaryOp,
    CConstant,
    CExpression,
    CFunction,
    CFunctionCall,
    CIndexedVariable,
    CStructField,
    CUnaryOp,
    CVariable,
    CVariableField,
)
from angr.sim_type import SimType
from angr.sim_variable import SimMemoryVariable, SimStackVariable, SimTemporaryVariable, SimVariable
from pyqodeng.core import api, modes, panels
from pyqodeng.core.api.syntax_highlighter import COLOR_SCHEME_KEYS
from PySide6.QtCore import QEvent, Qt
from PySide6.QtGui import QAction, QKeySequence
from PySide6.QtWidgets import QApplication, QInputDialog, QLineEdit, QMenu

from angrmanagement.ui.dialogs.rename_node import RenameNode
from angrmanagement.ui.dialogs.retype_node import RetypeNode
from angrmanagement.ui.dialogs.xref import XRefDialog
from angrmanagement.ui.documents.qcodedocument import QCodeDocument
from angrmanagement.ui.menus.menu import Menu
from angrmanagement.ui.widgets.qccode_highlighter import FORMATS, QCCodeHighlighter

if TYPE_CHECKING:
    from PySide6.QtGui import QTextDocument

    from angrmanagement.ui.views.code_view import CodeView


class ColorSchemeIDA(api.ColorScheme):
    """
    An IDA-like color scheme.
    """

    def __init__(self) -> None:
        super().__init__("default")
        for k, v in FORMATS.items():
            if k in COLOR_SCHEME_KEYS:
                self.formats[COLOR_SCHEME_KEYS[k]] = v
                self.formats[k] = v


class QCCodeEdit(api.CodeEdit):
    """
    A subclass of pyqodeng's CodeEdit, specialized to handle the kinds of textual interaction expected of the pseudocode
    view. You will typically interact with this class as code_view.textedit.
    """

    def __init__(self, code_view) -> None:
        self._palette_updating = False
        self.line_number_panel: panels.LineNumberPanel | None = None
        self.folding_panel: panels.FoldingPanel | None = None
        self.code_highlighter: QCCodeHighlighter | None = None
        self.color_scheme: api.ColorScheme | None = None

        super().__init__(create_default_actions=True)
        self._code_view: CodeView = code_view
        self.create_panels()

        self.modes.append(modes.SymbolMatcherMode())

        self.setTabChangesFocus(False)
        self.setReadOnly(True)

        self.constant_actions = []
        self.operator_actions = []
        self.variable_actions = []
        self.call_actions = []
        self.default_actions = []
        self.function_name_actions = []
        self.action_rename_node = None
        self.action_xref = None
        self._selected_node = None

        self._initialize_context_menus()

        # but we don't need some of the actions
        self.remove_action(self.action_undo)
        self.remove_action(self.action_redo)
        self.remove_action(self.action_cut)
        self.remove_action(self.action_paste)
        self.remove_action(self.action_duplicate_line)
        self.remove_action(self.action_swap_line_up)
        self.remove_action(self.action_swap_line_down)

    def create_panels(self):
        self.line_number_panel = panels.LineNumberPanel()
        self.folding_panel = panels.FoldingPanel()
        self.panels.append(self.line_number_panel)
        self.panels.append(self.folding_panel)
        self.line_number_panel.setVisible(True)
        self.folding_panel.setVisible(True)

    def remove_panels(self):
        if self.line_number_panel is not None:
            self.line_number_panel.setVisible(False)
            self.panels.remove(self.line_number_panel.name)
            self.line_number_panel = None
        if self.folding_panel is not None:
            self.folding_panel.setVisible(False)
            self.panels.remove(self.folding_panel.name)
            self.folding_panel = None

    def node_under_cursor(self):
        doc: QTextDocument = self.document()
        if not isinstance(doc, QCodeDocument):
            # this is not the qcodedocument that the decompiler generates. this means the pseudocode view is empty
            return None

        # determine the current status
        cursor = self.textCursor()
        pos = cursor.position()
        current_node = doc.get_node_at_position(pos)
        if current_node is not None:
            return current_node
        else:
            # nothing is under the cursor
            return None

    def get_context_menu(self):
        if self.document() is None:
            return QMenu()

        # TODO: Anything in sel?

        # Get the highlighted item
        under_cursor = self.node_under_cursor()

        mnu = QMenu()
        self._selected_node = None
        if isinstance(under_cursor, CConstant):
            self._selected_node = under_cursor
            mnu.addActions(self.constant_actions)
        if isinstance(under_cursor, CBinaryOp | CUnaryOp):
            # operator in selection
            self._selected_node = under_cursor
            mnu.addActions(self.operator_actions)
        if (
            isinstance(under_cursor, CFunctionCall)
            and "vex_block_addr" in under_cursor.tags
            and "ins_addr" in under_cursor.tags
        ):
            # function call in selection
            self._selected_node = under_cursor
            mnu.addActions(self.call_actions)
        if isinstance(under_cursor, CVariable | CIndexedVariable | CVariableField | CStructField):
            # variable in selection
            self._selected_node = under_cursor
            mnu.addActions(self.variable_actions)
        if isinstance(under_cursor, CFunction):
            # decompiled function name in selection
            self._selected_node = under_cursor
            mnu.addActions(self.function_name_actions)
            for entry in self.workspace.plugins.build_context_menu_functions(
                [self.instance.kb.functions[under_cursor.name]]
            ):
                Menu.translate_element(mnu, entry)
        else:
            mnu.addActions(self.default_actions)

        for entry in self.workspace.plugins.build_context_menu_node(under_cursor):
            Menu.translate_element(mnu, entry)

        for action in self.llm_actions:
            mnu.addAction(action)

        return mnu

    @property
    def workspace(self):
        return self._code_view.workspace if self._code_view is not None else None

    @property
    def instance(self):
        return self._code_view.instance if self._code_view is not None else None

    def event(self, event):
        """
        Reimplemented to capture the Tab key pressed/released event.

        :param event:
        :return:
        """

        if event.type() == QEvent.ShortcutOverride and event.key() == Qt.Key_Tab:
            event.accept()
            return True
        if event.type() == QEvent.KeyPress and event.key() == Qt.Key_Tab:
            return self.keyPressEvent(event)

        return super().event(event)

    def get_closest_insaddr(self, node, expr=None) -> int | None:
        addr = (getattr(node, "tags", None) or {}).get("ins_addr", None)
        if addr is None:
            if expr:
                addr = self.get_src_to_inst()
            else:
                pos = self.textCursor().position()
                # find until the next newline
                while (
                    self.document().characterAt(pos) not in ("\n", "\u2029") and pos < self.document().characterCount()
                ):  # qt WHAT are you doing
                    pos += 1
                prior_pos = pos - 1  # use the location before the newline, otherwise we get the node of the next line
                node = self.document().get_stmt_node_at_position(prior_pos)
                addr = (getattr(node, "tags", None) or {}).get("ins_addr", None)

        return addr

    def get_src_to_inst(self) -> int:
        """
        Uses the current cursor position, which is in a code view, and gets the
        corresponding instruction address that is associated to the code.
        Returns the start of the function if unable to calculate.

        :return: int (address of inst)
        """

        # get the Qt document
        doc: QCodeDocument = self.document()

        # get the current position of the cursor
        cursor = self.textCursor()
        pos = cursor.position()

        # get the node at the associated cursor position
        current_node = doc.get_stmt_node_at_position(pos)

        if (
            current_node is not None
            and hasattr(current_node, "tags")
            and current_node.tags is not None
            and "ins_addr" in current_node.tags
        ):
            asm_ins_addr = current_node.tags["ins_addr"]

        else:
            # the top of the function decompiled
            asm_ins_addr = self._code_view.function.addr

        return asm_ins_addr

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()
        xkey = key
        if isinstance(xkey, int):
            xkey = Qt.Key(xkey)
        if modifiers & Qt.ShiftModifier:
            xkey |= Qt.Key.Key_Shift
        if modifiers & Qt.ControlModifier:
            xkey |= Qt.Key.Key_Control
        if modifiers & Qt.AltModifier:
            xkey |= Qt.Key.Key_Alt
        if modifiers & Qt.MetaModifier:
            xkey |= Qt.Key.Key_Meta
        sequence = QKeySequence(xkey)
        mnu = self.get_context_menu()
        for item in mnu.actions():
            if item.shortcut().matches(sequence) == QKeySequence.SequenceMatch.ExactMatch:
                item.activate(QAction.ActionEvent.Trigger)
                return True

        if key in (Qt.Key_Slash, Qt.Key_Question):
            self.comment(
                expr=event.modifiers() & Qt.KeyboardModifier.ShiftModifier == Qt.KeyboardModifier.ShiftModifier
            )
            return True
        if key == Qt.Key_Minus and QApplication.keyboardModifiers() & Qt.KeyboardModifier.ControlModifier != 0:
            self.zoomOut()
        if key == Qt.Key_Equal and QApplication.keyboardModifiers() & Qt.KeyboardModifier.ControlModifier != 0:
            self.zoomIn()

        if self._code_view.keyPressEvent(event):
            return True

        saved_mode = self.textInteractionFlags()
        if key in (
            Qt.Key.Key_Left,
            Qt.Key.Key_Right,
            Qt.Key.Key_Up,
            Qt.Key.Key_Down,
            Qt.Key.Key_PageDown,
            Qt.Key.Key_PageUp,
            Qt.Key.Key_Home,
            Qt.Key.Key_End,
        ):
            self.setTextInteractionFlags(saved_mode | Qt.TextInteractionFlag.TextEditable)
        result = super().keyPressEvent(event)
        self.setTextInteractionFlags(saved_mode)
        return result

    def paintEvent(self, e) -> None:
        saved_mode = self.textInteractionFlags()
        self.setTextInteractionFlags(saved_mode | Qt.TextInteractionFlag.TextEditable)
        super().paintEvent(e)
        self.setTextInteractionFlags(saved_mode)

    def setDocument(self, document) -> None:
        super().setDocument(document)

        self.color_scheme = ColorSchemeIDA()
        self.code_highlighter = QCCodeHighlighter(self.document(), color_scheme=self.color_scheme)
        self.modes.append(self.code_highlighter)
        self.syntax_highlighter.fold_detector = api.CharBasedFoldDetector()

    def setPalette(self, palette, **kwargs) -> None:
        super().setPalette(palette, **kwargs)
        if self._palette_updating:
            return
        self._palette_updating = True
        # re-create panels to apply the new palette
        self.remove_panels()
        self.create_panels()
        # re-generate color scheme because FORMATS has changed
        if self.color_scheme is not None:
            self.color_scheme = ColorSchemeIDA()
        if self.code_highlighter is not None:
            self.code_highlighter.refresh_editor(self.color_scheme)
            self.rehighlight()
        self._palette_updating = False

    #
    # Actions
    #
    # pylint: disable=unused-argument
    def rename_node(self, *args, node=None) -> None:  # pylint: disable=unused-argument
        n = node if node is not None else self._selected_node
        if not isinstance(n, CVariable | CFunction | CFunctionCall | CStructField | SimType):
            return
        if isinstance(n, CVariable) and isinstance(n.variable, SimTemporaryVariable):
            # unsupported right now..
            return
        dialog = RenameNode(code_view=self._code_view, node=n, func=self._code_view.function)
        dialog.exec_()

    def xref_node(self, *args, node=None) -> None:  # pylint: disable=unused-argument
        n = node if node is not None else self._selected_node
        if not isinstance(n, CVariable | CFunction | CFunctionCall):
            return

        if isinstance(n, CFunction | CFunctionCall):
            addr = n.addr if isinstance(n, CFunction) else n.callee_func.addr
            dialog = XRefDialog(
                addr=addr,
                xrefs_manager=self.instance.project.kb.xrefs,
                dst_addr=addr,
                instance=self.instance,
                disassembly_view=self._code_view,
                parent=self._code_view,
            )
        else:
            addr = self.get_closest_insaddr(n)
            func_addr_or_global = (
                "global"
                if isinstance(n.variable, SimMemoryVariable) and not isinstance(n.variable, SimStackVariable)
                else None
            )
            dialog = XRefDialog(
                addr=addr,
                variable_manager=self._code_view.variable_manager(func_addr=func_addr_or_global),
                variable=n.variable,
                instance=self.instance,
                disassembly_view=self._code_view,
                parent=self._code_view,
            )
        dialog.exec_()

    def retype_node(self, *args, node=None, node_type=None) -> None:  # pylint: disable=unused-argument
        if node is None:
            node = self._selected_node
        if not isinstance(node, CVariable | CFunction | CFunctionCall | CStructField):
            return
        if isinstance(node, CVariable) and isinstance(node.variable, SimTemporaryVariable):
            # unsupported right now..
            return
        dialog = RetypeNode(self.instance, code_view=self._code_view, node=node, node_type=node_type)
        dialog.exec_()

        new_node_type = dialog.new_type
        if new_node_type is not None and self._code_view is not None and node is not None:
            if isinstance(node, CFunction):
                self._code_view.function.prototype = new_node_type.with_arch(self.instance.project.arch)
                self._code_view.codegen.am_event(event="retype_function", node=node)
                return

            # need workspace for altering callbacks of changes
            variable_kb = self._code_view.codegen._variable_kb
            # specify the type
            new_node_type = new_node_type.with_arch(self.instance.project.arch)
            variable_kb.variables[self._code_view.function.addr].set_variable_type(
                node.variable,
                new_node_type,
                all_unified=True,
                mark_manual=True,
            )

            self._code_view.codegen.am_event(event="retype_variable", node=node, variable=node.variable)

    def comment(self, expr: bool = False, node=None) -> None:
        addr = self.get_closest_insaddr(node, expr=expr)
        if addr is None:
            return

        try:
            cdict = self._code_view.codegen.expr_comments if expr else self._code_view.codegen.stmt_comments
        except AttributeError:
            return
        text = cdict.get(addr, "")

        text, ok = QInputDialog.getText(
            self._code_view, "Expression Comment" if expr else "Statement Comment", "", QLineEdit.EchoMode.Normal, text
        )

        if ok:
            exists = addr in cdict
            if text:
                # callback
                self.workspace.plugins.handle_comment_changed(addr, "", text, not exists, True)
                cdict[addr] = text
            else:
                if exists:
                    # callback
                    self.workspace.plugins.handle_comment_changed(addr, cdict[addr], "", False, True)
                    del cdict[addr]

            self._code_view.codegen.am_event()

    def toggle_struct(self) -> None:
        node = self._selected_node
        if not isinstance(node, CVariable):
            return
        if isinstance(node.variable, SimVariable):
            ident = node.variable.ident
            if ident in self._code_view.vars_must_struct:
                self._code_view.vars_must_struct.remove(ident)
            else:
                self._code_view.vars_must_struct.add(ident)
            # decompile again
            self._code_view.decompile()

    def collapse_expr(self) -> None:
        if hasattr(self._selected_node, "collapsed"):
            self._selected_node.collapsed = True
            self._code_view.codegen.am_event()

    def expand_expr(self) -> None:
        if hasattr(self._selected_node, "collapsed"):
            self._selected_node.collapsed = False
            self._code_view.codegen.am_event()

    def hex_constant(self) -> None:
        if hasattr(self._selected_node, "fmt_hex"):
            self._selected_node.fmt_hex ^= True
            self._code_view.codegen.am_event()

    def char_constant(self) -> None:
        if hasattr(self._selected_node, "fmt_char"):
            self._selected_node.fmt_char ^= True
            self._code_view.codegen.am_event()

    def neg_constant(self) -> None:
        if hasattr(self._selected_node, "fmt_neg"):
            self._selected_node.fmt_neg ^= True
            self._code_view.codegen.am_event()

    def float_constant(self) -> None:
        if hasattr(self._selected_node, "fmt_float"):
            self._selected_node.fmt_float ^= True
            self._code_view.codegen.am_event()

    def double_constant(self) -> None:
        if hasattr(self._selected_node, "fmt_double"):
            self._selected_node.fmt_double ^= True
            self._code_view.codegen.am_event()

    def convert_to_ite_expr(self) -> None:
        node = self._selected_node
        if not isinstance(node, CExpression):
            return
        ailexpr = self._code_view.codegen.cnode2ailexpr.get(node, None)
        if ailexpr is None:
            return

        # which statement?
        addr = self.get_closest_insaddr(node)
        if addr is None:
            return

        cache = self.instance.kb.decompilations[(self._code_view.function.addr, "pseudocode")]
        if cache.ite_exprs is None:
            cache.ite_exprs = set()
        cache.ite_exprs.add((addr, ailexpr))
        self._code_view.decompile(regen_clinic=False)

    def swap_binop_operands(self) -> None:
        node = self._selected_node
        if not isinstance(node, CBinaryOp):
            return
        ailexpr = self._code_view.codegen.cnode2ailexpr.get(node, None)
        if ailexpr is None:
            return
        if not isinstance(ailexpr, BinaryOp):
            return

        op = ailexpr.op
        negated_op = op if op in {"CmpEQ", "CmpNE"} else BinaryOp.COMPARISON_NEGATION.get(op, None)
        if negated_op is None:
            return

        # which statement?
        addr = self.get_closest_insaddr(node)
        if addr is None:
            return

        cache = self.instance.kb.decompilations[(self._code_view.function.addr, "pseudocode")]
        if cache.binop_operators is None:
            cache.binop_operators = {}
        op_desc = OpDescriptor(
            ailexpr.vex_block_addr if hasattr(ailexpr, "vex_block_addr") else None,
            ailexpr.vex_stmt_idx if hasattr(ailexpr, "vex_stmt_idx") else None,
            addr,
            op,
        )

        existing_op_desc_removed = False
        if negated_op in {"CmpEQ", "CmpNE"} or negated_op != op:
            # remove existing descriptor if we are swapping the same binop expression twice
            existing_op_desc = OpDescriptor(
                ailexpr.vex_block_addr if hasattr(ailexpr, "vex_block_addr") else None,
                ailexpr.vex_stmt_idx if hasattr(ailexpr, "vex_stmt_idx") else None,
                addr,
                negated_op,
            )
            if existing_op_desc in cache.binop_operators:
                del cache.binop_operators[existing_op_desc]
                existing_op_desc_removed = True

        if not existing_op_desc_removed:
            cache.binop_operators[op_desc] = negated_op
        self._code_view.decompile(regen_clinic=False)

    def expr2armasm(self) -> None:
        def _assemble(expr, expr_addr) -> str:
            return converter.assemble(expr, self._code_view.function.addr, expr_addr)

        def _find_loads(expr) -> list:
            if isinstance(expr, Load):
                return [expr]
            elif isinstance(expr, Op):
                if isinstance(expr, UnaryOp):
                    return _find_loads(expr.operand)
                else:
                    loads = []
                    for operand in expr.operands:
                        loads += _find_loads(operand)
                    return loads
            else:
                return []

        node = self._selected_node
        # figure out where we are
        if not isinstance(node, CVariable | CIndexedVariable | CVariableField | CStructField):
            return

        doc: QCodeDocument = self.document()
        cursor = self.textCursor()
        pos = cursor.position()
        current_node = doc.get_stmt_node_at_position(pos)

        if current_node is None:
            return

        ins_addr = current_node.tags.get("ins_addr", None)
        if ins_addr is None:
            return

        # traverse the stored clinic graph to find the AIL block
        cache = self.instance.kb.decompilations[(self._code_view.function.addr, "pseudocode")]
        the_node = None
        for node in cache.clinic.graph.nodes():
            if node.addr <= ins_addr < node.addr + node.original_size:
                the_node = node
                break

        if the_node is None:
            return

        converter = self.workspace.plugins.get_plugin_instance_by_name("AIL2ARM32")

        lst = []
        for stmt in the_node.statements:
            if isinstance(stmt, Assignment):
                loads = _find_loads(stmt.src)
                for load in loads:
                    asm = _assemble(load, load.ins_addr)
                    lst.append((str(stmt), str(load), asm))
            elif isinstance(stmt, Store):
                loads = _find_loads(stmt.data)
                for load in loads:
                    asm = _assemble(load, load.ins_addr)
                    lst.append((str(stmt), str(load), asm))

        # format text
        text = f"The AIL block:\n{str(the_node)}\n\n"
        for stmt, expr, asm in lst:
            text += f"Statement: {stmt}\n"
            text += f"Expression: {expr}\n"
            text += f"Assembly:\n{asm}\n\n"

        converter.display_output(text)

    #
    # Private methods
    #

    @staticmethod
    def _separator():
        sep = QAction()
        sep.setSeparator(True)
        return sep

    def _llm_refine_all(self) -> None:
        self.workspace.llm_refine_all()

    def _llm_suggest_var_names(self) -> None:
        self.workspace.llm_suggest_variable_names()

    def _llm_suggest_func_name(self) -> None:
        self.workspace.llm_suggest_function_name()

    def _llm_suggest_var_types(self) -> None:
        self.workspace.llm_suggest_variable_types()

    def _initialize_context_menus(self) -> None:
        base_actions = [
            self._separator(),
            self.action_copy,
            self.action_select_all,
        ]

        self.action_rename_node = QAction("Rename...", self)
        self.action_rename_node.triggered.connect(self.rename_node)
        self.action_rename_node.setShortcut(QKeySequence("N"))
        self.action_xref = QAction("Xrefs...", self)
        self.action_xref.triggered.connect(self.xref_node)
        self.action_xref.setShortcut(QKeySequence("X"))
        self.action_retype_node = QAction("Specify a new type", self)
        self.action_retype_node.triggered.connect(self.retype_node)
        self.action_retype_node.setShortcut(QKeySequence("Y"))
        self.action_toggle_struct = QAction("Toggle &struct/array")
        self.action_toggle_struct.triggered.connect(self.toggle_struct)
        self.action_collapse_expr = QAction("Collapse expression", self)
        self.action_collapse_expr.triggered.connect(self.collapse_expr)
        self.action_expand_expr = QAction("Expand expression", self)
        self.action_expand_expr.triggered.connect(self.expand_expr)
        self.action_hex = QAction("Toggle hex", self)
        self.action_hex.triggered.connect(self.hex_constant)
        self.action_hex.setShortcut(QKeySequence("H"))
        self.action_neg = QAction("Toggle negative", self)
        self.action_neg.triggered.connect(self.neg_constant)
        self.action_neg.setShortcut(QKeySequence("_"))
        self.action_char = QAction("Toggle char", self)
        self.action_char.triggered.connect(self.char_constant)
        self.action_char.setShortcut(QKeySequence("R"))
        self.action_float = QAction("Toggle float", self)
        self.action_float.triggered.connect(self.float_constant)
        self.action_double = QAction("Toggle double", self)
        self.action_double.triggered.connect(self.double_constant)
        self.action_to_ite_expr = QAction("Create a ternary expression")
        self.action_to_ite_expr.triggered.connect(self.convert_to_ite_expr)
        self.action_swap_binop_operands = QAction("Swap operands")
        self.action_swap_binop_operands.triggered.connect(self.swap_binop_operands)

        expr_actions = [
            self.action_to_ite_expr,
            self.action_swap_binop_operands,
            self.action_collapse_expr,
            self.action_expand_expr,
        ]

        self.action_asmgen = QAction("Expression -> ARM THUMB assembly...")
        self.action_asmgen.triggered.connect(self.expr2armasm)

        self.variable_actions = [
            self.action_rename_node,
            self.action_retype_node,
            self.action_toggle_struct,
            self.action_asmgen,
            self.action_xref,
        ]

        self.function_name_actions = [self.action_rename_node, self.action_xref, self.action_retype_node]

        self.constant_actions = [
            self.action_hex,
            self.action_neg,
            self.action_char,
            self.action_float,
            self.action_double,
        ]

        self.call_actions = [self.action_rename_node, self.action_xref]

        self.constant_actions += base_actions + expr_actions
        self.operator_actions += base_actions + expr_actions
        self.variable_actions += base_actions + expr_actions
        self.function_name_actions += base_actions
        self.call_actions += base_actions + expr_actions
        self.default_actions += base_actions

        # LLM actions
        self.action_llm_refine_all = QAction("AI: Refine All", self)
        self.action_llm_refine_all.triggered.connect(self._llm_refine_all)
        self.action_llm_suggest_var_names = QAction("AI: Suggest Variable Names", self)
        self.action_llm_suggest_var_names.triggered.connect(self._llm_suggest_var_names)
        self.action_llm_suggest_func_name = QAction("AI: Suggest Function Name", self)
        self.action_llm_suggest_func_name.triggered.connect(self._llm_suggest_func_name)
        self.action_llm_suggest_var_types = QAction("AI: Suggest Variable Types", self)
        self.action_llm_suggest_var_types.triggered.connect(self._llm_suggest_var_types)
        self.llm_actions = [
            self._separator(),
            self.action_llm_refine_all,
            self.action_llm_suggest_var_names,
            self.action_llm_suggest_func_name,
            self.action_llm_suggest_var_types,
        ]



from __future__ import annotations

from typing import TYPE_CHECKING

from angr.sim_type import SimType, SimTypePointer

if TYPE_CHECKING:
    from angr.knowledge_plugins import Function


def type2str(ty: SimType | None) -> str:
    """
    Convert a SimType instance to a string that can be displayed.

    :param ty:      The SimType instance, or None if it's for void.
    :return:        A string.
    """

    if ty is None:
        return "void"
    if isinstance(ty, SimTypePointer):
        return f"{type2str(ty.pts_to)}*"
    if ty.label:
        return ty.label
    return ty.c_repr()


def function_prototype_str(func: Function) -> str:
    if func.prototype is None:
        return func.name

    # Type of the return value
    s = ""
    rt = type2str(func.prototype.returnty)
    s += rt + " "

    # function name
    s += func.demangled_name
    s += "("

    # arguments
    for i, arg_type in enumerate(func.prototype.args):
        type_str = type2str(arg_type)

        if func.prototype.arg_names and i < len(func.prototype.arg_names):
            arg_name = func.prototype.arg_names[i]
        else:
            arg_name = f"arg_{i}"

        s += type_str + " " + arg_name

        if i < len(func.prototype.args) - 1:
            # splitter
            s += ","

    s += ")"
    return s



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QAbstractTableModel, QSize, Qt
from PySide6.QtWidgets import (
    QAbstractItemView,
    QButtonGroup,
    QDialog,
    QGroupBox,
    QHBoxLayout,
    QHeaderView,
    QPushButton,
    QRadioButton,
    QTableView,
    QVBoxLayout,
)

if TYPE_CHECKING:
    from angr.flirt import FlirtSignature

    from angrmanagement.data.instance import Instance


class QSignatureMatchesTableModel(QAbstractTableModel):
    """
    Table model for signature matches.
    """

    Headers = ["Function Address", "Function Name", "Conflict"]
    COL_ADDRESS = 0
    COL_NAME = 1
    COL_CONFLICT = 2

    def __init__(self, matches: dict[int, str], instance: Instance) -> None:
        super().__init__()
        self.matches = matches
        self.instance = instance
        self.match_list = sorted(matches.items())  # Sort by address

    def rowCount(self, parent=None) -> int:
        return len(self.match_list)

    def columnCount(self, parent=None) -> int:
        return len(self.Headers)

    def headerData(self, section: int, orientation, role=None):
        if role != Qt.ItemDataRole.DisplayRole:
            return None
        if orientation == Qt.Orientation.Horizontal and section < len(self.Headers):
            return self.Headers[section]
        return None

    def data(self, index, role=None):
        if not index.isValid():
            return None
        if role == Qt.ItemDataRole.DisplayRole:
            row = index.row()
            col = index.column()
            if row >= len(self.match_list):
                return None

            addr, name = self.match_list[row]
            if col == self.COL_ADDRESS:
                return f"{addr:#x}"
            elif col == self.COL_NAME:
                return name
            elif col == self.COL_CONFLICT:
                # Check if function has a non-default name
                func = self.instance.kb.functions.get(addr)
                if func and not func.is_default_name:
                    # Check if the name matches what we're trying to apply (not a conflict)
                    if func.name == name:
                        return ""
                    # Check if the name is actually a "generic" name that shouldn't count as a conflict
                    name_lower = func.name.lower()
                    if not name_lower.startswith(("unknown", "unresolved", "sub_")):
                        return "✓"
                return ""
        return None


class QSignatureMatchesTable(QTableView):
    """
    Table widget for displaying signature matches.
    """

    def __init__(self, matches: dict[int, str], instance: Instance, parent=None) -> None:
        super().__init__(parent)
        self.instance = instance
        self.matches = matches

        self.horizontalHeader().setVisible(True)
        self.verticalHeader().setVisible(False)
        self.verticalHeader().setDefaultSectionSize(20)

        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollMode.ScrollPerPixel)

        self.model_obj = QSignatureMatchesTableModel(matches, instance)
        self.setModel(self.model_obj)

        hheader = self.horizontalHeader()
        hheader.setSectionResizeMode(QSignatureMatchesTableModel.COL_ADDRESS, QHeaderView.ResizeMode.ResizeToContents)
        hheader.setSectionResizeMode(QSignatureMatchesTableModel.COL_NAME, QHeaderView.ResizeMode.Stretch)
        hheader.setSectionResizeMode(QSignatureMatchesTableModel.COL_CONFLICT, QHeaderView.ResizeMode.ResizeToContents)

        self.doubleClicked.connect(self._on_double_click)

    def _on_double_click(self, index) -> None:
        """Jump to the function when double-clicked."""
        row = index.row()
        if row >= len(self.model_obj.match_list):
            return
        addr, _name = self.model_obj.match_list[row]
        # Jump to the function address
        from angrmanagement.logic import GlobalInfo

        GlobalInfo.main_window.workspace.jump_to(addr)
        # Close the dialog
        self.window().accept()


class SignatureMatchesDialog(QDialog):
    """
    Dialog displaying matched functions for a signature.
    """

    def __init__(
        self, sig: FlirtSignature, matches: dict[int, str], instance: Instance, signature_mgr, parent=None
    ) -> None:
        super().__init__(parent)
        self.sig = sig
        self.matches = matches
        self.instance = instance
        self.signature_mgr = signature_mgr

        self.setWindowTitle(f"Signature Matches: {sig.sig_name}")
        self._init_widgets()

        self.setWindowFlags(
            self.windowFlags() & ~Qt.WindowType.WindowContextHelpButtonHint | Qt.WindowType.WindowCloseButtonHint
        )
        self.setMinimumSize(self.sizeHint())

    def sizeHint(self):
        return QSize(600, 400)

    def is_conflict(self, addr: int, name: str) -> bool:
        func = self.instance.kb.functions.get(addr)
        if func and not func.is_default_name:
            if func.name == name:
                return False
            name_lower = func.name.lower()
            return not name_lower.startswith(("unknown", "unresolved", "sub_"))
        return False

    def _init_widgets(self) -> None:
        table = QSignatureMatchesTable(self.matches, self.instance, self)

        # Count conflicts
        conflict_count = 0
        total_matches = len(self.matches)
        for addr, name in self.matches.items():
            if self.is_conflict(addr, name):
                conflict_count += 1

        # Conflict resolution radio buttons
        conflict_group = QGroupBox(f"Conflict Resolution ({conflict_count}/{total_matches} conflicts)")
        conflict_layout = QHBoxLayout()

        self.overwrite_radio = QRadioButton("Overwrite")
        self.ignore_radio = QRadioButton("Ignore")
        self.overwrite_radio.setChecked(True)  # Default to Overwrite

        self.conflict_button_group = QButtonGroup()
        self.conflict_button_group.addButton(self.overwrite_radio)
        self.conflict_button_group.addButton(self.ignore_radio)

        conflict_layout.addWidget(self.overwrite_radio)
        conflict_layout.addWidget(self.ignore_radio)
        conflict_layout.addStretch()
        conflict_group.setLayout(conflict_layout)

        # Apply button
        apply_button = QPushButton("Apply Functions")
        apply_button.clicked.connect(self._on_apply_functions)

        # Main layout
        layout = QVBoxLayout()
        layout.addWidget(table)
        layout.addWidget(conflict_group)
        layout.addWidget(apply_button)
        self.setLayout(layout)

    def _on_apply_functions(self) -> None:
        """Handle Apply Functions button click."""
        ignore_addresses = set()

        if self.ignore_radio.isChecked():
            for addr, name in self.matches.items():
                if self.is_conflict(addr, name):
                    ignore_addresses.add(addr)

        self.signature_mgr.apply_signatures([self.sig], dry_run=False, ignore_addresses=ignore_addresses)
        self.accept()



from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from angrmanagement.data.instance import Instance


def extract_first_paragraph_from_docstring(desc: str) -> str:
    desc = textwrap.dedent(desc).strip()  # Remove docstring indent
    desc = desc[0 : desc.find("\n\n")]  # First paragraph
    desc = desc.replace("\n", " ")  # Unwrap
    return desc


class AnalysesConfiguration:
    """
    Configuration for a sequence of analyses.
    """

    def __init__(self, analyses: Sequence[AnalysisConfiguration]) -> None:
        self.analyses: Sequence[AnalysisConfiguration] = analyses

    def __len__(self) -> int:
        return len(self.analyses)

    def __iter__(self):
        return iter(self.analyses)

    def __getitem__(self, key: int | str):
        if isinstance(key, int):
            return self.analyses[key]
        return self.by_name(key)

    def by_name(self, name: str) -> AnalysisConfiguration:
        for a in self.analyses:
            if a.name == name:
                return a
        raise KeyError(name)


class AnalysisConfiguration:
    """
    Configuration for an analysis.
    """

    def __init__(self, instance: Instance) -> None:
        self.instance = instance
        self.enabled: bool = False
        self.name: str = ""
        self.display_name: str = ""
        self.description: str = "Description not available"
        self.options: Mapping[str, AnalysisOption] = {}

    def __getitem__(self, key: str):
        return self.options[key]

    def to_dict(self):
        """
        Return dictionary with configuration for this option.
        """
        o = {}
        self.update_dict(o)
        return o

    def update_dict(self, out: dict[str, Any]) -> None:
        """
        Update dictionary `out` with configuration for this option.
        """
        for o in self.options.values():
            if getattr(o, "enabled", None) is False:
                continue
            o.update_dict(out)

    def get_main_obj_size(self) -> int:
        main_obj_size = 0
        if self.instance.project.loader.main_object is not None:
            main_obj = self.instance.project.loader.main_object
            if main_obj.segments:
                for seg in main_obj.segments:
                    if seg.is_executable:
                        main_obj_size += seg.memsize
            if main_obj_size == 0 and main_obj.sections:
                # fall back to sections
                for sec in main_obj.sections:
                    if sec.is_executable:
                        main_obj_size += sec.memsize
        return main_obj_size


class AnalysisOption:
    """
    Configurable option for an analysis.
    """

    def __init__(self, name: str, display_name: str, tooltip: str) -> None:
        self.name: str = name
        self.display_name: str = display_name
        self.tooltip: str = tooltip

    def update_dict(self, out: dict[str, Any]) -> None:
        """
        Update dictionary `out` with configuration for this option.
        """


class PrimitiveAnalysisOption(AnalysisOption):
    """
    Configurable option for an analysis, with a fundamental type (e.g. bool)
    """

    def __init__(self, name: str, description: str, default: Any, tooltip: str) -> None:
        super().__init__(name, description, tooltip)
        self.default: Any = default
        self.value: Any = default

    def update_dict(self, out: dict[str, Any]) -> None:
        """
        Update `out` dictionary with configuration for this option.
        """
        out[self.name] = self.value


class BoolAnalysisOption(PrimitiveAnalysisOption):
    """
    Boolean option for an analysis.
    """

    def __init__(self, name: str, description: str, default: bool = False, tooltip: str = "") -> None:
        super().__init__(name, description, default, tooltip)


class StringAnalysisOption(PrimitiveAnalysisOption):
    """
    String option for an analysis.
    """

    def __init__(self, name: str, description: str, default: str = "", tooltip: str = "") -> None:
        super().__init__(name, description, default, tooltip)

    @property
    def enabled(self):
        return bool(self.value)


class IntAnalysisOption(PrimitiveAnalysisOption):
    """
    Integer option for an analysis.
    """

    def __init__(
        self,
        name: str,
        description: str,
        default: int = 0,
        tooltip: str = "",
        minimum: int | None = None,
        maximum: int | None = None,
    ) -> None:
        super().__init__(name, description, default, tooltip)
        self.minimum_value = minimum
        self.maximum_value = maximum


class ChoiceAnalysisOption(PrimitiveAnalysisOption):
    """
    A multi-value choice.
    """

    def __init__(
        self, name: str, description: str, choices: Mapping[Any, str], default: Any, tooltip: str = ""
    ) -> None:
        super().__init__(name, description, default, tooltip)
        self.choices = choices



"""
Test cases for MainWindow keyboard shortcuts and event filters.
"""
# pylint: disable=no-self-use,missing-class-docstring

from __future__ import annotations

import time
import unittest
from unittest.mock import MagicMock, patch

from common import AngrManagementTestCase
from PySide6.QtCore import QEvent, Qt
from PySide6.QtGui import QKeyEvent, QWindow
from PySide6.QtWidgets import QWidget

from angrmanagement.ui.main_window import DockShortcutEventFilter, ShiftShiftEventFilter


class TestDockShortcutEventFilter(AngrManagementTestCase):
    """Test DockShortcutEventFilter for Ctrl+Shift+P shortcut."""

    def setUp(self):
        super().setUp()
        self.event_filter = DockShortcutEventFilter(self.main)

    def test_ctrl_shift_p_triggers_command_palette(self):
        """Test that Ctrl+Shift+P opens the command palette."""
        with patch.object(self.main, "show_command_palette") as mock_show:
            mock_widget = QWidget()
            key_event = QKeyEvent(
                QKeyEvent.Type.KeyPress,
                Qt.Key.Key_P,
                Qt.KeyboardModifier.ControlModifier | Qt.KeyboardModifier.ShiftModifier,
            )

            result = self.event_filter.eventFilter(mock_widget, key_event)

            assert result is True
            mock_show.assert_called_once_with(mock_widget)
            mock_widget.deleteLater()

    def test_other_key_combinations_not_handled(self):
        """Test that other key combinations are not handled by the filter."""
        mock_widget = QWidget()
        key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_P, Qt.KeyboardModifier.ControlModifier)

        result = self.event_filter.eventFilter(mock_widget, key_event)

        assert result is False
        mock_widget.deleteLater()

    def test_non_keypress_events_not_handled(self):
        """Test that non-KeyPress events are not handled."""
        mock_widget = QWidget()
        event = QEvent(QEvent.Type.MouseButtonPress)

        result = self.event_filter.eventFilter(mock_widget, event)

        assert result is False
        mock_widget.deleteLater()


class TestShiftShiftEventFilter(AngrManagementTestCase):
    """Test ShiftShiftEventFilter for double-Shift shortcut."""

    def setUp(self):
        super().setUp()
        self.event_filter = ShiftShiftEventFilter(self.main)

    def test_double_shift_triggers_goto_palette(self):
        """Test that pressing Shift twice quickly opens goto palette."""
        with patch.object(self.main, "show_goto_palette") as mock_show:
            mock_window = MagicMock(spec=QWindow)
            mock_window.modality.return_value = Qt.WindowModality.NonModal
            mock_widget = QWidget()

            # Create key events for Shift press
            key_event1 = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_Shift, Qt.KeyboardModifier.NoModifier)
            key_event2 = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_Shift, Qt.KeyboardModifier.NoModifier)

            # First Shift press - simulates Qt event propagation (QWindow -> QWidget)
            self.event_filter.eventFilter(mock_window, key_event1)
            self.event_filter.eventFilter(mock_widget, key_event1)

            # Second Shift press - should trigger goto palette
            self.event_filter.eventFilter(mock_window, key_event2)
            result = self.event_filter.eventFilter(mock_widget, key_event2)

            assert result is True
            mock_show.assert_called_once_with(mock_widget)
            mock_widget.deleteLater()

    def test_single_shift_does_not_trigger(self):
        """Test that a single Shift press does not trigger goto palette."""
        with patch.object(self.main, "show_goto_palette") as mock_show:
            mock_window = MagicMock(spec=QWindow)
            mock_window.modality.return_value = Qt.WindowModality.NonModal
            mock_widget = QWidget()

            key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_Shift, Qt.KeyboardModifier.NoModifier)

            # Single Shift press (QWindow -> QWidget)
            self.event_filter.eventFilter(mock_window, key_event)
            result = self.event_filter.eventFilter(mock_widget, key_event)

            assert result is False
            mock_show.assert_not_called()
            mock_widget.deleteLater()

    def test_shift_presses_outside_timeout_reset(self):
        """Test that Shift presses outside timeout window reset the counter."""
        with patch.object(self.main, "show_goto_palette") as mock_show:
            mock_window = MagicMock(spec=QWindow)
            mock_window.modality.return_value = Qt.WindowModality.NonModal
            mock_widget = QWidget()

            key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_Shift, Qt.KeyboardModifier.NoModifier)

            # First Shift press
            self.event_filter.eventFilter(mock_window, key_event)
            self.event_filter.eventFilter(mock_widget, key_event)

            # Simulate time passing beyond the timeout without real sleeping
            self.event_filter._last_press_time = time.time() - (self.event_filter.timeout_secs + 0.1)

            # Second Shift press (after simulated timeout)
            key_event2 = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_Shift, Qt.KeyboardModifier.NoModifier)
            self.event_filter.eventFilter(mock_window, key_event2)
            result = self.event_filter.eventFilter(mock_widget, key_event2)

            # Should not trigger because timeout expired
            assert result is False
            mock_show.assert_not_called()
            mock_widget.deleteLater()

    def test_other_keys_reset_counter(self):
        """Test that pressing other keys resets the Shift counter."""
        with patch.object(self.main, "show_goto_palette") as mock_show:
            mock_window = MagicMock(spec=QWindow)
            mock_window.modality.return_value = Qt.WindowModality.NonModal
            mock_widget = QWidget()

            shift_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_Shift, Qt.KeyboardModifier.NoModifier)
            other_key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_A, Qt.KeyboardModifier.NoModifier)

            # First Shift press
            self.event_filter.eventFilter(mock_window, shift_event)
            self.event_filter.eventFilter(mock_widget, shift_event)

            # Press another key
            self.event_filter.eventFilter(mock_window, other_key_event)
            self.event_filter.eventFilter(mock_widget, other_key_event)

            # Second Shift press (counter should have reset)
            shift_event2 = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_Shift, Qt.KeyboardModifier.NoModifier)
            self.event_filter.eventFilter(mock_window, shift_event2)
            result = self.event_filter.eventFilter(mock_widget, shift_event2)

            assert result is False
            mock_show.assert_not_called()
            mock_widget.deleteLater()


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

__version__ = "9.2.207"


try:
    # make sure qtpy (which is used in PyQodeNG.core) is using PySide6
    import os

    os.environ["QT_API"] = "pyside6"
    import qtpy  # noqa
except ImportError:
    # qtpy is not installed
    pass


def boot(project=None, block=False):
    if block:
        _boot(project)
        # noreturn
    else:
        import queue
        import threading

        response = queue.Queue()
        t = threading.Thread(target=_boot, args=(project, response))
        t.start()
        return response.get()


def _boot(project=None, response=None):
    import glob
    import os

    from PySide6.QtCore import QThread
    from PySide6.QtGui import QFontDatabase, QIcon
    from PySide6.QtWidgets import QApplication

    app = QApplication()
    app.setApplicationDisplayName("angr")
    app.setApplicationName("angr")
    from .consts import FONT_LOCATION, IMG_LOCATION

    icon_location = os.path.join(IMG_LOCATION, "angr.png")
    QApplication.setWindowIcon(QIcon(icon_location))

    from .config import Conf

    Conf.attempt_importing_initial_config()
    from .logic import GlobalInfo
    from .ui.awesome_tooltip_event_filter import QAwesomeTooltipEventFilter
    from .ui.css import refresh_theme  # import .ui after showing the splash screen since it's going to take time
    from .ui.main_window import MainWindow

    refresh_theme()
    for font_file in glob.glob(os.path.join(FONT_LOCATION, "*.ttf")):
        QFontDatabase.addApplicationFont(font_file)
    Conf.init_font_config()
    Conf.connect("ui_default_font", app.setFont, True)

    # install the global tooltip filter
    app.installEventFilter(QAwesomeTooltipEventFilter(app))

    GlobalInfo.gui_thread = QThread.currentThread()
    main_window = MainWindow(app=app)
    if project is not None:
        main_window.workspace.main_instance.project.am_obj = project
        main_window.workspace.main_instance.project.am_event()
        cfg = project.kb.cfgs.get_most_accurate()
        if cfg is not None:
            main_window.workspace.main_instance.cfg.am_obj = cfg
            main_window.workspace.main_instance.cfg.am_event()
    else:
        main_window.show_welcome_dialog()
    if response:
        response.put(main_window.workspace)
    app.exec_()



"""
Test cases for DataDepView.
"""

from __future__ import annotations

import unittest
from unittest.mock import MagicMock, patch

from common import ProjectOpenTestCase
from PySide6.QtCore import Qt
from PySide6.QtGui import QKeyEvent
from PySide6.QtWidgets import QDialog

from angrmanagement.ui.dialogs.data_dep_graph_search import QDataDepGraphSearch
from angrmanagement.ui.views.data_dep_view import DataDepView


class TestDataDepView(ProjectOpenTestCase):
    """Test DataDepView functionality."""

    def setUp(self):
        super().setUp()

        self.data_dep_view = DataDepView(self.workspace, "center", self.instance)

        self._exec_called = False
        self._dialog_instance = None

        def mock_exec(dialog_self, *_args, **_kwargs):
            self._exec_called = True
            self._dialog_instance = dialog_self
            return QDialog.DialogCode.Accepted

        self._exec_patcher = patch("PySide6.QtWidgets.QDialog.exec_", mock_exec)
        self._exec_patcher.start()
        self.addCleanup(self._exec_patcher.stop)

    def test_search_dialog_opens_with_ctrl_f(self):
        """Test that DataDepView opens search dialog with Ctrl+F and uses exec_() (modal)."""
        mock_graph_widget = MagicMock()

        with patch.object(
            type(self.data_dep_view), "graph_widget", new_callable=lambda: property(lambda self: mock_graph_widget)
        ):
            self._exec_called = False
            key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_F, Qt.KeyboardModifier.ControlModifier)
            self.data_dep_view.keyPressEvent(key_event)

            assert self._exec_called, "Search dialog should use .exec_() (modal)"
            assert isinstance(self._dialog_instance, QDataDepGraphSearch)
            self._dialog_instance.close()


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from PySide6.QtCore import Qt
from PySide6.QtGui import QTextCharFormat

from angrmanagement.config import Conf

if TYPE_CHECKING:
    from PySide6.QtWidgets import QGraphicsSceneMouseEvent

    from angrmanagement.logic.disassembly import InfoDock


class BlockTreeNodeOptions:
    """
    Various options to control display of QBlockCodeObj's
    """

    show_conditional_jump_targets: bool = True
    show_variables: bool = True
    show_variable_identifiers: bool = True


class BlockTreeNode:
    """
    Renders a generic "code" object and handles display related events.
    Instances of this class mirror an AST structure, with references in the
    `obj` property to whatever object should be displayed. Leaf nodes will add
    text to the display document during render, which is handled at the top
    level by `QBlockCode`.
    """

    obj: Any
    infodock: InfoDock
    parent: Any
    options: BlockTreeNodeOptions
    span: tuple[int, int] | None
    subobjs: list[BlockTreeNode | str]
    _fmt_current: QTextCharFormat
    display_address: bool
    top_margin_lines: int
    bottom_margin_lines: int

    def __init__(
        self,
        obj: Any,
        infodock: InfoDock,
        parent: Any,
        options: BlockTreeNodeOptions | None = None,
        display_address: bool = True,
        top_margin_lines: int = 0,
        bottom_margin_lines: int = 0,
    ) -> None:
        self.obj = obj
        self.infodock = infodock
        self.parent = parent
        self.options = options or BlockTreeNodeOptions()
        self.span = None
        self.subobjs = []
        self.display_address = display_address
        self._fmt_current = None  # type: ignore
        self.top_margin_lines = top_margin_lines
        self.bottom_margin_lines = bottom_margin_lines
        self.update_style()
        self.create_subobjs(obj)

    @staticmethod
    def fmt() -> QTextCharFormat:
        """
        Get text char formatting for this object
        """
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_node_mnemonic_color)
        return fmt

    @property
    def selection_key(self) -> Any:
        return None

    def update_style(self) -> None:
        """
        Updates current rendering style before draw
        """
        self._fmt_current = self.fmt()
        if self.should_highlight():
            self._fmt_current.setBackground(Conf.disasm_view_operand_highlight_color)
            # self._fmt_current.setFontWeight(QFont.Weight.Bold)
        else:
            self._fmt_current.setBackground(Qt.GlobalColor.transparent)
            # self._fmt_current.setFontWeight(QFont.Weight.Normal)

    def should_highlight(self) -> bool:
        """
        Determine whether this object should be drawn with highlight
        """
        selected = self.infodock.selected_block_tree_node
        selection_key = self.selection_key
        r = (selected is not None) and (
            (selection_key is not None and selected == selection_key)
            or selected is self
            or (isinstance(selected, BlockTreeNode) and selected.obj is self.obj)
        )
        return r

    def create_subobjs(self, obj) -> None:
        """
        Initialize any display subobjects for this object
        """

    def recreate_subobjs(self) -> None:
        self.subobjs.clear()
        self.create_subobjs(self.obj)

    def update(self) -> None:
        """
        Update self and parent objects
        """
        self.parent.update()

    def render_to_doc(self, cursor) -> None:
        """
        Add each subobject to the document
        """
        self.update_style()
        self.recreate_subobjs()
        span_min = cursor.position()
        for obj in self.subobjs:
            if isinstance(obj, str):
                cursor.insertText(obj, self._fmt_current)
            else:
                obj.render_to_doc(cursor)
        span_max = cursor.position()
        self.span = (span_min, span_max)

    def hit_test(self, pos: int) -> bool:
        """
        Determine whether a character offset falls within the span of this object
        """
        if self.span is None:
            return False
        return self.span[0] <= pos < self.span[1]

    def get_hit_obj(self, pos: int) -> BlockTreeNode | None:
        """
        Find the leaf node for a given character offset
        """
        if not self.hit_test(pos):
            return None
        for obj in self.subobjs:
            if not isinstance(obj, str):
                hit = obj.get_hit_obj(pos)
                if hit is not None:
                    return hit
        return self

    def _add_subobj(self, obj: BlockTreeNode | str) -> None:
        """
        Add display object `obj` to the list of subobjects
        """
        self.subobjs.append(obj)

    def add_text(self, text: str) -> None:
        """
        Add a text leaf
        """
        self._add_subobj(text)

    def add_newline(self, cnt: int = 1) -> None:
        """
        Add a newline leaf
        """
        self._add_subobj("\n" * cnt)

    def add_variable(self, var) -> None:
        self._add_subobj(QVariableObj(var, self.infodock, parent=self, options=self.options))

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:  # pylint: disable=unused-argument
        selection_key = self.selection_key
        if selection_key is None:
            selection_key = self

        if event.button() == Qt.MouseButton.LeftButton:
            self.infodock.toggle_block_tree_node_selection(selection_key)
        elif event.button() == Qt.MouseButton.RightButton:
            self.infodock.select_block_tree_node(selection_key)
            self.infodock.disasm_view.show_context_menu_for_selected_object()

    def mouseDoubleClickEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        pass

    @property
    def should_highlight_line(self):
        return any(obj.should_highlight_line for obj in self.subobjs if isinstance(obj, BlockTreeNode))


class QVariableObj(BlockTreeNode):
    """
    Renders a variable
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_variable_label_color)
        return fmt

    def create_subobjs(self, obj) -> None:
        ident = "<%s>" % (obj.ident if obj.ident else "") if self.options.show_variable_identifiers else ""
        self.add_text(obj.name + ident)



from __future__ import annotations

from .calling_convention_recovery import CallingConventionRecoveryConfiguration, CallingConventionRecoveryJob
from .cfg_generation import CFGAnalysisConfiguration, CFGGenerationJob
from .code_tagging import CodeTaggingConfiguration, CodeTaggingJob
from .ddg_generation import DDGGenerationJob
from .decompile_function import DecompileFunctionJob
from .deobfuscation import (
    APIDeobfuscationConfiguration,
    APIDeobfuscationJob,
    StringDeobfuscationConfiguration,
    StringDeobfuscationJob,
)
from .dependency_analysis import DependencyAnalysisJob
from .flirt_signature_recognition import FlirtAnalysisConfiguration, FlirtSignatureRecognitionJob
from .job import Job
from .llm_preload import LLMPreloadCalleesJob
from .llm_refine import LLMRefineJob
from .loading import LoadBinaryJob
from .prototype_finding import PrototypeFindingJob
from .simgr_explore import SimgrExploreJob
from .simgr_step import SimgrStepJob
from .variable_recovery import VariableRecoveryConfiguration, VariableRecoveryJob
from .vfg_generation import VFGGenerationJob

__all__ = [
    "APIDeobfuscationConfiguration",
    "APIDeobfuscationJob",
    "StringDeobfuscationConfiguration",
    "StringDeobfuscationJob",
    "CallingConventionRecoveryConfiguration",
    "CallingConventionRecoveryJob",
    "CFGAnalysisConfiguration",
    "CFGGenerationJob",
    "CodeTaggingConfiguration",
    "CodeTaggingJob",
    "DDGGenerationJob",
    "DecompileFunctionJob",
    "DependencyAnalysisJob",
    "FlirtAnalysisConfiguration",
    "FlirtSignatureRecognitionJob",
    "Job",
    "LLMPreloadCalleesJob",
    "LLMRefineJob",
    "PrototypeFindingJob",
    "SimgrExploreJob",
    "SimgrStepJob",
    "VariableRecoveryConfiguration",
    "VariableRecoveryJob",
    "VFGGenerationJob",
    "LoadBinaryJob",
]



from __future__ import annotations

import math
from typing import TYPE_CHECKING

from PySide6.QtGui import QColor

from .trace_statistics import TraceStatistics

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace


class MultiTrace:
    HIT_COLOR = QColor(0x00, 0x99, 0x00, 0x60)
    MISS_COLOR = QColor(0xEE, 0xEE, 0xEE)
    FUNCTION_NOT_VISITED_COLOR = QColor(0x99, 0x00, 0x00, 0x20)
    BUCKET_COLORS = [
        QColor(0xEF, 0x65, 0x48, 0x20),
        QColor(0xFC, 0x8D, 0x59, 0x60),
        QColor(0xFD, 0xBB, 0x84, 0x60),
        QColor(0xFD, 0xD4, 0x9E, 0x60),
    ]

    def __init__(self, workspace: Workspace) -> None:
        self.workspace = workspace
        self._traces_summary = []
        self._traces = {}
        self.function_info = {}
        self.is_active_tab = False
        self.addr_color_map = {}

    def add_trace(self, trace, base_addr):
        traceStats = TraceStatistics(self.workspace, trace, base_addr)
        self._traces[trace["id"]] = traceStats
        self._traces_summary.extend(traceStats.mapped_trace)
        return traceStats

    def get_hit_miss_color(self, addr: int):
        if addr in self.addr_color_map:
            return self.addr_color_map[addr]
        else:
            return MultiTrace.MISS_COLOR

    def get_percent_color(self, func):
        addr = func.addr
        if addr in self.addr_color_map:
            return self.addr_color_map[addr]
        return None

    def get_coverage(self, func):
        if func.addr not in self.function_info:
            self._calc_function_info(func)
        return self.function_info[func.addr]["coverage"]

    def get_any_trace(self, addr: int):
        for trace in self._traces.values():
            if addr in trace["trace"]:
                return trace["trace"]

        return None

    def get_all_trace_ids(self):
        return self._traces.keys()

    def get_input_id_for_trace_id(self, trace_id):
        if trace_id not in self._traces:
            self.workspace.log(f"ERROR - trace id {trace_id} not present in multitrace")
            return None
        trace = self._traces[trace_id]
        return trace.input_id

    def get_trace_with_id(self, trace_id):
        if trace_id not in self._traces:
            self.workspace.log(f"ERROR - trace id {trace_id} not present in multitrace")
            return None
        return self._traces[trace_id]

    def clear_heatmap(self) -> None:
        self._make_addr_map([])

    def reload_heatmap(self, targets) -> None:
        addrs_of_interest = []
        for trace_id in targets:
            if trace_id not in self._traces:
                self.workspace.log(f"{trace_id} not found in traces")
                continue
            addr_list = self._traces[trace_id].mapped_trace
            addrs_of_interest.extend(addr_list)
        self._make_addr_map(addrs_of_interest)

    def _make_addr_map(self, addrs_of_interest) -> None:
        # TODO: Probably exists a more efficient way to generate this mapping
        self.addr_color_map.clear()
        hit_map = {}
        for addr in addrs_of_interest:
            if addr not in hit_map:
                hit_map[addr] = 0
            hit_map[addr] += 1

        buckets = {}
        for addr, count in hit_map.items():
            if count not in buckets:
                buckets[count] = []
            buckets[count].append(addr)

        strata_size = math.floor(len(set(addrs_of_interest)) / 9)

        total = 0
        bucket_counts = sorted(buckets.keys())
        for count in bucket_counts:
            addrs = buckets[count]
            density = 50 + math.floor(total / strata_size) * 20
            color = QColor(0xFF, 0xFF, 0x30, density)
            for addr in addrs:
                self.addr_color_map[addr] = color
            total += len(addrs)

    def _calc_function_info(self, func) -> None:
        blocks = list(func.block_addrs)
        hit_count = 0

        for block_addr in blocks:
            # hexstr_addr = hex(block)
            if block_addr in self._traces_summary:
                hit_count += 1

        if hit_count == 0:
            self.function_info[func.addr] = {"color": MultiTrace.FUNCTION_NOT_VISITED_COLOR, "coverage": 0}
        elif hit_count == len(blocks):
            self.function_info[func.addr] = {"color": MultiTrace.HIT_COLOR, "coverage": 100}
        else:
            hit_percent = (hit_count / len(blocks)) * 100
            bucket_size = 100 / len(MultiTrace.BUCKET_COLORS)
            bucket_pos = math.floor(hit_percent / bucket_size)
            self.function_info[func.addr] = {"color": MultiTrace.BUCKET_COLORS[bucket_pos], "coverage": hit_percent}



# pylint:disable=no-self-use,unused-argument
from __future__ import annotations

import sys
from enum import Enum, auto
from typing import Any, ClassVar, Generic, TypeVar

from PySide6.QtCore import (
    QAbstractItemModel,
    QEvent,
    QModelIndex,
    QPersistentModelIndex,
    QRect,
    QSortFilterProxyModel,
    Qt,
    Signal,
)
from PySide6.QtGui import QBrush, QColor, QFont, QMouseEvent, QPalette, QPen, QPixmap
from PySide6.QtWidgets import (
    QApplication,
    QColorDialog,
    QComboBox,
    QDoubleSpinBox,
    QFileDialog,
    QFontDialog,
    QHBoxLayout,
    QLineEdit,
    QSpinBox,
    QSplitter,
    QStyle,
    QStyledItemDelegate,
    QStyleOptionViewItem,
    QTextEdit,
    QToolButton,
    QTreeView,
    QVBoxLayout,
    QWidget,
)

from angrmanagement.ui.icons import icon


class PropertyType(Enum):
    """
    PropertyItem types.
    """

    INVALID = auto()
    GROUP = auto()
    INT = auto()
    TEXT = auto()
    COMBO = auto()
    FLOAT = auto()
    COLOR = auto()
    BOOL = auto()
    FONT = auto()
    FILE = auto()


T = TypeVar("T")


class PropertyItem:
    """
    Base PropertyItem. Don't use this directly.
    """

    type: ClassVar[PropertyType] = PropertyType.INVALID

    def __init__(self, name: str, description: str = "", **kwargs):
        assert self.type != PropertyType.INVALID
        self.name = name
        self.description = description
        self.parent: PropertyItem | None = None
        self.children: list[PropertyItem] = []
        for prop in kwargs.pop("children", []):
            self.addChild(prop)
        self.extra: dict[str, Any] = kwargs

    def childCount(self):
        return len(self.children)

    def child(self, row):
        if 0 <= row < len(self.children):
            return self.children[row]
        return None

    def row(self):
        if self.parent:
            return self.parent.children.index(self)
        return 0

    def addChild(self, child):
        child.parent = self
        self.children.append(child)


class GroupPropertyItem(PropertyItem):
    """
    Property group item.
    """

    type: ClassVar[PropertyType] = PropertyType.GROUP


class ValuePropertyItem(PropertyItem, Generic[T]):
    """
    Generic value property.
    """

    initial_value: T
    value: T

    def __init__(self, name: str, value: T, description: str = "", readonly: bool = False, **kwargs):
        super().__init__(name, description=description, **kwargs)
        self.initial_value = value
        self.value = value
        self.readonly = readonly

    @property
    def is_value_modified(self) -> bool:
        return self.value != self.initial_value


class BoolPropertyItem(ValuePropertyItem[bool]):
    """
    Boolean value property.
    """

    type: ClassVar[PropertyType] = PropertyType.BOOL

    def __init__(self, name: str, value: bool, description: str = "", **kwargs):
        super().__init__(name, value, description=description, **kwargs)


class ComboPropertyItem(ValuePropertyItem[Any]):
    """
    Combo property.
    """

    type: ClassVar[PropertyType] = PropertyType.COMBO
    choices: dict[Any, str]

    def __init__(self, name: str, value: Any, choices: list[Any] | dict[Any, str], description: str = "", **kwargs):
        super().__init__(name, value, description=description, **kwargs)
        assert isinstance(choices, list | dict)
        self.choices = {c: c for c in choices} if isinstance(choices, list) else choices


class IntPropertyItem(ValuePropertyItem[int]):
    """
    Integer property.
    """

    type: ClassVar[PropertyType] = PropertyType.INT

    def __init__(
        self,
        name: str,
        value: int,
        minimum: int = -(2**31),
        maximum: int = 2**31 - 1,
        description: str = "",
        show_hex: bool = False,
        **kwargs,
    ):
        super().__init__(name, value, description=description, **kwargs)
        self.minimum = minimum
        self.maximum = maximum
        self.show_hex = show_hex


class FloatPropertyItem(ValuePropertyItem[float]):
    """
    Real number property.
    """

    type: ClassVar[PropertyType] = PropertyType.FLOAT

    # FIXME: Float max, min
    def __init__(
        self,
        name: str,
        value: float,
        minimum: float = sys.float_info.min,
        maximum: float = sys.float_info.max,
        description: str = "",
        **kwargs,
    ):
        super().__init__(name, value, description=description, **kwargs)
        self.minimum = minimum
        self.maximum = maximum


class TextPropertyItem(ValuePropertyItem[str]):
    """
    Text property.
    """

    type: ClassVar[PropertyType] = PropertyType.TEXT


class FilePropertyItem(ValuePropertyItem[str]):
    """
    File path property.
    """

    type: ClassVar[PropertyType] = PropertyType.FILE


class ColorPropertyItem(ValuePropertyItem[QColor]):
    """
    Color property.
    """

    type: ClassVar[PropertyType] = PropertyType.COLOR


class FontPropertyItem(ValuePropertyItem[QFont]):
    """
    Font property.
    """

    type: ClassVar[PropertyType] = PropertyType.FONT


class PropertyModel(QAbstractItemModel):
    """
    Property tree model.
    """

    valueChanged = Signal(object, object)  # Emits (propertyItem, newValue)

    def __init__(self, root, parent=None):
        super().__init__(parent)
        self.rootItem = root

    def columnCount(self, parent=None):
        return 2

    def rowCount(self, parent=None):
        parentItem = parent.internalPointer() if parent and parent.isValid() else self.rootItem
        return parentItem.childCount()

    def index(self, row, column, parent=None):
        parent = parent or QModelIndex()
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        parentItem = parent.internalPointer() if parent.isValid() else self.rootItem
        childItem = parentItem.child(row)
        if childItem:
            return self.createIndex(row, column, childItem)
        return QModelIndex()

    def parent(self, index=None):  # type: ignore
        index = index or QModelIndex()
        if not index.isValid():
            return QModelIndex()
        childItem = index.internalPointer()
        parentItem = childItem.parent
        if parentItem == self.rootItem or parentItem is None:
            return QModelIndex()
        return self.createIndex(parentItem.row(), 0, parentItem)

    def data(self, index: QModelIndex | QPersistentModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if not index.isValid():
            return None
        item = index.internalPointer()
        column = index.column()

        if item.type == PropertyType.GROUP:
            if role == Qt.ItemDataRole.BackgroundRole:
                return QBrush(QApplication.palette().color(QPalette.ColorRole.Mid))
            if role == Qt.ItemDataRole.FontRole:
                font = QFont()
                font.setBold(True)
                return font

        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            if column == 0:
                return item.name
            elif column == 1:
                match item.type:
                    case PropertyType.GROUP:
                        return ""
                    case PropertyType.TEXT:
                        return item.value
                    case PropertyType.INT:
                        return str(item.value) if item.show_hex is False else hex(item.value)
                    case PropertyType.FLOAT:
                        return str(item.value)
                    case PropertyType.BOOL:
                        return ""  # Checkbox is rendered.
                    case PropertyType.COLOR:
                        return item.value.name()
                    case PropertyType.FONT:
                        return f"{item.value.family()}, {item.value.pointSize()}pt"
                    case PropertyType.FILE:
                        return item.value if item.value is not None else ""
                    case PropertyType.COMBO:
                        return item.choices[item.value]
                    case _:
                        raise NotImplementedError(item.type)

        if (
            role == Qt.ItemDataRole.DecorationRole
            and column == 1
            and item.type == PropertyType.COLOR
            and isinstance(item.value, QColor)
        ):
            pixmap = QPixmap(16, 16)
            pixmap.fill(item.value)
            return pixmap

        if role == Qt.ItemDataRole.CheckStateRole and column == 1 and item.type == PropertyType.BOOL:
            return Qt.CheckState.Checked if item.value else Qt.CheckState.Unchecked

        # Show modified properties in bold.
        if role == Qt.ItemDataRole.FontRole and item.type != PropertyType.GROUP and item.value != item.initial_value:
            font = QFont()
            font.setBold(True)
            return font

        return None

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return "Property" if section == 0 else "Value"
        return None

    def flags(self, index):
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        item = index.internalPointer()
        flags = Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled
        if index.column() == 1 and item.type != PropertyType.GROUP:
            flags |= Qt.ItemFlag.ItemIsEditable
        if item.type == PropertyType.BOOL:
            flags |= Qt.ItemFlag.ItemIsUserCheckable
        return flags

    def setData(
        self, index: QModelIndex | QPersistentModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole
    ) -> bool:
        if not index.isValid():
            return False
        item = index.internalPointer()
        if getattr(item, "readonly", False):
            return False
        if index.column() == 1:
            match item.type:
                case PropertyType.INT:
                    try:
                        item.value = int(value)
                    except ValueError:
                        return False
                case PropertyType.FLOAT:
                    try:
                        item.value = float(value)
                    except ValueError:
                        return False
                case PropertyType.BOOL:
                    if value in (Qt.CheckState.Checked, Qt.CheckState.Unchecked):
                        item.value = value == Qt.CheckState.Checked
                    else:
                        item.value = bool(value)
                case PropertyType.COLOR:
                    if isinstance(value, QColor):
                        item.value = value
                    else:
                        return False
                case PropertyType.FONT:
                    if isinstance(value, QFont):
                        item.value = value
                    else:
                        return False
                case PropertyType.FILE:
                    if isinstance(value, str):
                        item.value = value
                    else:
                        return False
                case _:
                    item.value = value
            self.dataChanged.emit(index, index, [role])
            name_idx = self.index(index.row(), 0, index.parent())
            self.dataChanged.emit(name_idx, name_idx, [Qt.ItemDataRole.DisplayRole])
            self.valueChanged.emit(item, item.value)
            return True
        return False


def for_all_model_rows(model, callback) -> None:
    """Iterate over all rows in the model."""
    stack = [QModelIndex()]
    while stack:
        parent = stack.pop()
        callback(parent)

        for row in range(model.rowCount(parent)):
            index = model.index(row, 0, parent)
            if not index.isValid():
                continue
            stack.append(index)


def get_level(index) -> int:
    """Returns the hierarchical depth level of the given index."""
    level = 0
    while index.parent().isValid():
        index = index.parent()
        level += 1
    return level


class PropertyDelegate(QStyledItemDelegate):
    """
    Property tree delegate.
    """

    @staticmethod
    def _get_source_item(index):
        # If the index is from a proxy model, map to source.
        model = index.model()
        if hasattr(model, "mapToSource"):
            source_index = model.mapToSource(index)
            return source_index.internalPointer()
        return index.internalPointer()

    def createEditor(
        self, parent: QWidget, option: QStyleOptionViewItem, index: QModelIndex | QPersistentModelIndex
    ) -> QWidget:
        item = self._get_source_item(index)
        match item.type:
            case PropertyType.INT:
                if item.show_hex:
                    editor = QLineEdit(parent)
                else:
                    editor = QSpinBox(parent)
                    editor.setRange(item.minimum, item.maximum)
                if item.readonly:
                    editor.setReadOnly(True)
                return editor
            case PropertyType.FLOAT:
                editor = QDoubleSpinBox(parent)
                editor.setDecimals(item.extra.get("decimals", 2))
                editor.setRange(item.minimum, item.maximum)
                if item.readonly:
                    editor.setReadOnly(True)
                return editor
            case PropertyType.TEXT:
                editor = QLineEdit(parent)
                if item.readonly:
                    editor.setReadOnly(True)
                return editor
            case PropertyType.COMBO:
                editor = QComboBox(parent)
                for data, text in item.choices.items():
                    editor.addItem(text, data)
                return editor
            case PropertyType.BOOL | PropertyType.COLOR | PropertyType.FONT | PropertyType.FILE:
                return None  # type: ignore
            case _:
                pass
        return super().createEditor(parent, option, index)

    def setEditorData(self, editor, index):
        item = self._get_source_item(index)
        value = item.value
        match item.type:
            case PropertyType.INT:
                if item.show_hex:
                    assert isinstance(editor, QLineEdit)
                    editor.setText(hex(value))
                else:
                    assert isinstance(editor, QSpinBox)
                    editor.setValue(int(value))
            case PropertyType.FLOAT:
                editor.setValue(float(value))
            case PropertyType.TEXT:
                editor.setText(str(value))
            case PropertyType.COMBO:
                idx = editor.findData(value)
                if idx >= 0:
                    editor.setCurrentIndex(idx)
            case _:
                super().setEditorData(editor, index)

    def setModelData(self, editor, model, index):
        item = self._get_source_item(index)
        match item.type:
            case PropertyType.INT:
                if item.show_hex:
                    assert isinstance(editor, QLineEdit)
                    try:
                        int_value = int(editor.text(), 16)
                    except ValueError:
                        int_value = item.value
                    model.setData(index, int_value, Qt.ItemDataRole.EditRole)
                else:
                    assert isinstance(editor, QSpinBox)
                    model.setData(index, editor.value(), Qt.ItemDataRole.EditRole)
            case PropertyType.FLOAT:
                model.setData(index, editor.value(), Qt.ItemDataRole.EditRole)
            case PropertyType.TEXT:
                model.setData(index, editor.text(), Qt.ItemDataRole.EditRole)
            case PropertyType.COMBO:
                model.setData(index, editor.currentData(), Qt.ItemDataRole.EditRole)
            case _:
                super().setModelData(editor, model, index)

    def updateEditorGeometry(self, editor, option, index):
        if editor:
            editor.setGeometry(option.rect)

    def editorEvent(self, event, model, option, index):
        item = self._get_source_item(index)
        if isinstance(event, QMouseEvent):
            single_clicked_on_value = (
                event.type() == QEvent.Type.MouseButtonRelease
                and event.button() == Qt.MouseButton.LeftButton
                and index.column() == 1
            )
            double_clicked_on_value = (
                event.type() == QEvent.Type.MouseButtonDblClick
                and event.button() == Qt.MouseButton.LeftButton
                and index.column() == 1
            )

            match item.type:
                case PropertyType.COLOR:
                    if double_clicked_on_value:
                        current_color = item.value if isinstance(item.value, QColor) else QColor("white")
                        new_color = QColorDialog.getColor(
                            current_color,
                            option.widget,
                            "Select Color",
                            QColorDialog.ColorDialogOption.ShowAlphaChannel,
                        )
                        if new_color.isValid():
                            model.setData(index, new_color, Qt.ItemDataRole.EditRole)
                        return True
                    return False
                case PropertyType.FONT:
                    if double_clicked_on_value:
                        current_font = item.value if hasattr(item.value, "family") else QFont()
                        ok, font = QFontDialog.getFont(current_font, option.widget, "Select Font")
                        if ok:
                            model.setData(index, font, Qt.ItemDataRole.EditRole)
                        return True
                    return False
                case PropertyType.FILE:
                    if double_clicked_on_value:
                        file_path, _ = QFileDialog.getOpenFileName(option.widget, "Select File", "", "All Files (*)")
                        if file_path:
                            model.setData(index, file_path, Qt.ItemDataRole.EditRole)
                        return True
                    return False
                case PropertyType.BOOL:
                    if single_clicked_on_value:
                        new_value = not item.value
                        model.setData(
                            index,
                            Qt.CheckState.Checked if new_value else Qt.CheckState.Unchecked,
                            Qt.ItemDataRole.EditRole,
                        )
                        return True
                    return False
                case _:
                    pass
        return super().editorEvent(event, model, option, index)

    def paint(self, painter, option, index):
        item = self._get_source_item(index)
        widget = option.widget
        style = widget.style() if widget else QApplication.style()
        palette = widget.palette()
        grid_color = palette.color(palette.ColorRole.Mid)

        # Get full row width
        row_rect = QRect(option.rect)
        row_rect.setLeft(widget.viewport().rect().left())
        row_rect.setRight(widget.viewport().rect().right())

        opt = QStyleOptionViewItem(option)
        opt_rect: QRect = opt.rect  # type: ignore

        if index.column() == 0:
            # Draw group background color
            if item.type == PropertyType.GROUP:
                painter.fillRect(row_rect, index.data(Qt.ItemDataRole.BackgroundRole))

            # Draw row highlight manually to ensure full row coverage
            if opt.state & QStyle.State_Selected:  # type: ignore
                style.drawPrimitive(QStyle.PrimitiveElement.PE_PanelItemViewItem, option, painter, widget)

            # Adjust for gutter
            opt_rect.adjust(opt_rect.height(), 0, 0, 0)

            # Indent levels
            indentation_per_level = 15
            level = get_level(index)
            if level > 1:
                opt_rect.adjust((level - 1) * indentation_per_level, 0, 0, 0)

        super().paint(painter, opt, index)

        if index.column() == 0:
            # Draw expand control
            branch = QStyleOptionViewItem()
            branch.rect = QRect(row_rect.x(), opt_rect.y(), opt_rect.height(), opt_rect.height())  # type: ignore
            branch.state = option.state  # type: ignore
            style.drawPrimitive(QStyle.PrimitiveElement.PE_IndicatorBranch, branch, painter)

        # Draw horizontal grid line
        painter.save()
        pen = QPen(grid_color, 1, Qt.PenStyle.SolidLine)  # Set grid color and thickness
        painter.setPen(pen)
        painter.drawLine(row_rect.bottomLeft(), row_rect.bottomRight())
        painter.restore()


class PropertyFilterProxy(QSortFilterProxyModel):
    """
    A proxy model for filtering displayed property tree items.
    """

    def filterAcceptsRow(self, source_row, source_parent):
        index0 = self.sourceModel().index(source_row, 0, source_parent)
        if not index0.isValid():
            return False
        # Filter based on property name.
        filter_text = self.filterRegularExpression().pattern().lower()
        if not filter_text:
            return True
        prop_name = self.sourceModel().data(index0, Qt.ItemDataRole.DisplayRole)
        return prop_name and filter_text in prop_name.lower()


class PropertyTreeView(QTreeView):
    """
    QTreeView subclass to correctly handle expand/collapse clicks with custom delegate painter that adjusts
    indentation and expand control placement.
    """

    def __init__(self):
        super().__init__()

        # Disable expand/collapse controls. We'll handle it manually.
        self.setItemsExpandable(False)
        self.setStyleSheet("""
            QTreeView::item {
                padding: 2px;
            }
            QTreeView::branch {
                border-image: none;
                image: none;
            }
        """)

    def mouseDoubleClickEvent(self, event):
        index = self.indexAt(event.pos())
        if index.isValid() and index.column() == 0:
            self.setExpanded(index, not self.isExpanded(index))
            return
        super().mouseDoubleClickEvent(event)

    def mousePressEvent(self, event):
        """Intercept mouse events to properly handle clicks on the expand/collapse indicator."""

        index = self.indexAt(event.pos())  # Get the clicked index
        if not index.isValid():
            super().mousePressEvent(event)
            return

        indicator_rect = QRect(self.visualRect(index))
        indicator_rect.setLeft(self.viewport().rect().x())
        indicator_rect.setWidth(indicator_rect.height())

        # If click is within the indicator area, manually trigger expand/collapse
        if indicator_rect.contains(event.pos()):
            self.setExpanded(index, not self.isExpanded(index))
            return  # Prevent further event propagation

        # Otherwise, proceed with the normal event handling
        super().mousePressEvent(event)


class QPropertyEditor(QWidget):
    """
    Main widget with a text filter box, the main tree view, and a description box.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        search_icon = icon("search", color_role=QPalette.ColorRole.PlaceholderText)
        assert search_icon is not None

        tools_layout = QHBoxLayout()
        tools_layout.setContentsMargins(0, 0, 0, 0)
        tools_layout.setSpacing(3)

        self._filter_box = QLineEdit()
        self._filter_box.setClearButtonEnabled(True)
        self._filter_box.addAction(search_icon, QLineEdit.ActionPosition.LeadingPosition)
        self._filter_box.setPlaceholderText("Filter by name...")
        self._filter_box.textChanged.connect(self._on_filter_text_changed)
        tools_layout.addWidget(self._filter_box)

        docs_icon = icon("docs")
        assert docs_icon is not None

        self._desc_btn = QToolButton()
        self._desc_btn.setIcon(docs_icon)
        self._desc_btn.setCheckable(True)
        self._desc_btn.setChecked(True)
        tools_layout.addWidget(self._desc_btn)

        self._tree_view = PropertyTreeView()
        self._tree_view.setAutoScroll(False)
        self._tree_view.setAlternatingRowColors(True)
        self._tree_view.setItemDelegate(PropertyDelegate())
        self._tree_view.expandAll()
        self._tree_view.expanded.connect(self._on_item_expanded_or_collapsed)
        self._tree_view.collapsed.connect(self._on_item_expanded_or_collapsed)

        self._desc_box = QTextEdit()
        self._desc_box.setReadOnly(True)
        self._desc_btn.clicked.connect(lambda: self._desc_box.setVisible(self._desc_btn.isChecked()))

        splitter = QSplitter()
        splitter.setOrientation(Qt.Orientation.Vertical)
        splitter.addWidget(self._tree_view)
        splitter.addWidget(self._desc_box)
        splitter.setStretchFactor(0, 1)
        splitter.setStretchFactor(1, 0)
        splitter.setSizes([100, 75])

        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(3)
        main_layout.addLayout(tools_layout)
        main_layout.addWidget(splitter)

        self._updating_expansion: bool = False
        self._expansion_preference: set[PropertyItem] = set()

        self._proxy_model = PropertyFilterProxy()
        self._proxy_model.setFilterCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self._proxy_model.setRecursiveFilteringEnabled(True)

    def set_description_visible(self, visible: bool = True):
        self._desc_btn.setVisible(visible)
        self._desc_btn.setChecked(visible)
        self._desc_box.setVisible(visible)

    def _on_filter_text_changed(self, text: str):
        self._updating_expansion = True
        self._proxy_model.setFilterFixedString(text)
        self._updating_expansion = False
        self._restore_expansion_state()

    def _save_expansion_state(self):
        def on_node(index):
            if index.isValid() and self._tree_view.isExpanded(index):
                source_index = self._proxy_model.mapToSource(index)
                self._expansion_preference.add(source_index.internalPointer())

        self._expansion_preference.clear()
        for_all_model_rows(self._proxy_model, on_node)

    def _restore_expansion_state(self):
        def on_node(index):
            source_index = self._proxy_model.mapToSource(index)
            prop = source_index.internalPointer()
            if prop in self._expansion_preference:
                self._tree_view.setExpanded(index, True)

        self._updating_expansion = True
        for_all_model_rows(self._proxy_model, on_node)
        self._updating_expansion = False

    def _on_item_expanded_or_collapsed(self, _):
        if not self._updating_expansion:
            self._save_expansion_state()

    def _on_item_selected(self, current, _):
        if current.isValid():
            source_index = self._proxy_model.mapToSource(current)
            item = source_index.internalPointer()
            self._desc_box.setHtml(f"<b>{item.name}</b><br>{item.description}")
        else:
            self._desc_box.clear()

    def setModel(self, model):
        self._desc_box.clear()
        self._proxy_model.setSourceModel(model)
        self._tree_view.setModel(self._proxy_model)
        self._tree_view.selectionModel().currentChanged.connect(self._on_item_selected)
        self._tree_view.setIndentation(0)
        self._tree_view.expandAll()
        self._tree_view.setColumnWidth(0, 300)
